#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                    Documentation                    #
#-----------------------------------------------------#
""" The classification variant of the ConvNeXt Large architecture.

| Architecture Variable    | Value                      |
| ------------------------ | -------------------------- |
| Key in architecture_dict | "2D.ConvNeXtLarge"         |
| Input_shape              | (384x384)                  |
| Standardization          | None                       |

Recommended alternative `Input_shape` is 224x224 pixels.

!!! warning
     ConvNeXt models expect their inputs to be float or uint8 tensors of pixels with values in the [0-255] range.
     Standardization is applied inside the architecture.

???+ abstract "Reference - Implementation"
    [https://www.tensorflow.org/api_docs/python/tf/keras/applications/convnext](https://www.tensorflow.org/api_docs/python/tf/keras/applications/convnext) <br>

???+ abstract "Reference - Publication"
    Zhuang Liu, Hanzi Mao, Chao-Yuan Wu, Christoph Feichtenhofer, Trevor Darrell, Saining Xie.
    10 Jan 2022. A ConvNet for the 2020s.
    <br>
    [https://arxiv.org/abs/2201.03545](https://arxiv.org/abs/2201.03545)
"""
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
from tensorflow.keras.applications.convnext import ConvNeXtLarge as BaseModel
# Internal libraries
from aucmedi.neural_network.architectures import Architecture_Base

#-----------------------------------------------------#
#          Architecture class: ConvNeXtLarge          #
#-----------------------------------------------------#
class ConvNeXtLarge(Architecture_Base):
    #---------------------------------------------#
    #                Initialization               #
    #---------------------------------------------#
    def __init__(self, classification_head, channels, input_shape=(224, 224),
                 pretrained_weights=False):
        self.classifier = classification_head
        self.input = input_shape + (channels,)
        self.pretrained_weights = pretrained_weights

    #---------------------------------------------#
    #                Create Model                 #
    #---------------------------------------------#
    def create_model(self):
        # Get pretrained image weights from imagenet if desired
        if self.pretrained_weights : model_weights = "imagenet"
        else : model_weights = None

        # Obtain ResNet50 as base model
        base_model = BaseModel(include_top=False, weights=model_weights,
                               input_tensor=None, input_shape=self.input,
                               pooling=None)
        top_model = base_model.output

        # Add classification head
        model = self.classifier.build(model_input=base_model.input,
                                      model_output=top_model)

        # Return created model
        return model



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                    Documentation                    #
#-----------------------------------------------------#
""" Library of implemented Aggregate functions in AUCMEDI.

An Aggregate function can be passed to an ensemble and merges multiple class predictions
into a single prediction.

```
Assembled predictions encoded in a NumPy matrix with shape (N_models, N_classes).
Example: [[0.5, 0.4, 0.1],
          [0.4, 0.3, 0.3],
          [0.5, 0.2, 0.3]]
-> shape (3, 3)

Merged prediction encoded in a NumPy matrix with shape (1, N_classes).
Example: [[0.4, 0.3, 0.3]]
-> shape (1, 3)
```

???+ example "Example"
    ```python
    # Recommended: Apply an Ensemble like Augmenting (test-time augmentation) with Majority Vote
    preds = predict_augmenting(model, test_datagen, n_cycles=5, aggregate="majority_vote")

    # Manual: Apply an Ensemble like Augmenting (test-time augmentation) with Majority Vote
    from aucmedi.ensemble.aggregate import MajorityVote
    my_agg = MajorityVote()
    preds = predict_augmenting(model, test_datagen, n_cycles=5, aggregate=my_agg)
    ```

Aggregate functions are based on the abstract base class [Aggregate_Base][aucmedi.ensemble.aggregate.agg_base.Aggregate_Base],
which allows simple integration of custom aggregate methods for Ensemble.
"""
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# Import aggregate functions
from aucmedi.ensemble.aggregate.averaging_mean import AveragingMean
from aucmedi.ensemble.aggregate.averaging_median import AveragingMedian
from aucmedi.ensemble.aggregate.majority_vote import MajorityVote
from aucmedi.ensemble.aggregate.softmax import Softmax
from aucmedi.ensemble.aggregate.global_argmax import GlobalArgmax

#-----------------------------------------------------#
#       Access Functions to Aggregate Functions       #
#-----------------------------------------------------#
aggregate_dict = {"mean": AveragingMean,
                  "median": AveragingMedian,
                  "majority_vote": MajorityVote,
                  "softmax": Softmax,
                  "global_argmax": GlobalArgmax,
}
""" Dictionary of implemented Aggregate functions. """



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
#External libraries
import unittest
import tempfile
import os
from PIL import Image
import SimpleITK as sitk
import numpy as np
import random
import pandas as pd
#Internal libraries
from aucmedi.automl.block_train import block_train
from aucmedi.automl.block_pred import block_predict
from aucmedi.automl.block_eval import block_evaluate

#-----------------------------------------------------#
#          Unittest: AutoML Evaluation Block          #
#-----------------------------------------------------#
class AutoML_block_evaluate(unittest.TestCase):
    # Create random imaging and classification data
    @classmethod
    def setUpClass(self):
        np.random.seed(1234)
        # Initialize temporary directory
        self.tmp_data2D = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                                      suffix=".data2D")
        # Create RGB data
        for i in range(0, 25):
            img_rgb = np.random.rand(32, 32, 3) * 255
            imgRGB_pillow = Image.fromarray(img_rgb.astype(np.uint8))
            index = "sample_" + str(i) + ".png"
            path_sampleRGB = os.path.join(self.tmp_data2D.name, index)
            imgRGB_pillow.save(path_sampleRGB)

        # Create multi-class classification labels
        data = {}
        for i in range(0, 25):
            data["sample_" + str(i)] = np.random.randint(4)
        self.tmp_csv = tempfile.NamedTemporaryFile(mode="w",
                                                   prefix="tmp.aucmedi.",
                                                   suffix=".csv")
        df = pd.DataFrame.from_dict(data, orient="index", columns=["CLASS"])
        df.index.name = "SAMPLE"
        df.to_csv(self.tmp_csv.name, index=True, header=True)

        # Initialize temporary directory
        self.model_dir = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                                     suffix=".train")
        # Define config
        config = {
            "path_imagedir": self.tmp_data2D.name,
            "path_gt": self.tmp_csv.name,
            "path_modeldir": self.model_dir.name,
            "analysis": "minimal",
            "ohe": False,
            "three_dim": False,
            "epochs": 8,
            "batch_size": 4,
            "workers": 1,
            "architecture": "Vanilla"
        }
        # Run AutoML training block
        block_train(config)

        # Define config
        self.pred_path = tempfile.NamedTemporaryFile(mode="w",
                                                     prefix="tmp.aucmedi.",
                                                     suffix=".pred.csv")
        config = {
            "path_imagedir": self.tmp_data2D.name,
            "path_modeldir": self.model_dir.name,
            "path_pred": self.pred_path.name,
            "batch_size": 4,
            "workers": 1,
            "xai_method": None,
            "xai_directory": None,
        }
        # Run AutoML inference block
        block_predict(config)

    #-------------------------------------------------#
    #              Performance Evaluation             #
    #-------------------------------------------------#
    def test_eval_performance(self):
        # Define config
        config = {
            "path_imagedir": self.tmp_data2D.name,
            "path_gt": self.tmp_csv.name,
            "path_pred": self.pred_path.name,
            "path_evaldir": self.model_dir.name,
            "ohe": False,
        }
        # Run evaluation block
        block_evaluate(config)

        self.assertTrue(len(os.listdir(self.model_dir.name)) == 8)
        self.assertTrue(os.path.exists(os.path.join(self.model_dir.name,
                                       "plot.performance.barplot.png")))
        self.assertTrue(os.path.exists(os.path.join(self.model_dir.name,
                                      "plot.performance.confusion_matrix.png")))
        self.assertTrue(os.path.exists(os.path.join(self.model_dir.name,
                                      "plot.performance.roc.png")))
        self.assertTrue(os.path.exists(os.path.join(self.model_dir.name,
                                      "metrics.performance.csv")))



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
#External libraries
import unittest
import tempfile
import os
from PIL import Image
import numpy as np
#Internal libraries
from aucmedi import *
from aucmedi.xai import *
from aucmedi.xai.methods import *
from aucmedi.utils.visualizer import visualize_heatmap
from aucmedi.data_processing.io_loader import image_loader

#-----------------------------------------------------#
#              Unittest: Explainable AI               #
#-----------------------------------------------------#
class xaiTEST(unittest.TestCase):
    # Setup AUCMEDI pipeline
    @classmethod
    def setUpClass(self):
        np.random.seed(1234)
        # Initialize temporary directory
        self.tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                                    suffix=".data")
        # Create RGB data
        self.sampleList = []
        for i in range(0, 10):
            img_rgb = np.random.rand(32, 32, 3) * 255
            img_pillow = Image.fromarray(img_rgb.astype(np.uint8))
            index = "image.sample_" + str(i) + ".RGB.png"
            path_sample = os.path.join(self.tmp_data.name, index)
            img_pillow.save(path_sample)
            self.sampleList.append(index)
        # Create classification labels
        self.labels_ohe = np.zeros((10, 4), dtype=np.uint8)
        for i in range(0, 10):
            class_index = np.random.randint(0, 4)
            self.labels_ohe[i][class_index] = 1

        # Create Data Generator
        self.datagen = DataGenerator(self.sampleList,  self.tmp_data.name,
                                     labels=self.labels_ohe, resize=None,
                                     grayscale=False, batch_size=3)
        # Create Neural Network model
        self.model = NeuralNetwork(n_labels=4, channels=3, input_shape=(32,32),
                                    architecture="2D.Vanilla")
        # Compute predictions
        self.preds = self.model.predict(self.datagen)
        # Initialize testing image
        self.image = self.datagen[0][0][[0]]

    #-------------------------------------------------#
    #             XAI Functions: Decoder              #
    #-------------------------------------------------#
    def test_Decoder_argmax_output(self):
        imgs, hms = xai_decoder(self.datagen, self.model, preds=self.preds,
                                out_path=None)
        self.assertTrue(np.array_equal(imgs.shape, (10, 32, 32, 3)))
        self.assertTrue(np.array_equal(hms.shape, (10, 32, 32)))

    def test_Decoder_argmax_visualize(self):
        path_xai = os.path.join(self.tmp_data.name, "xai")
        xai_decoder(self.datagen, self.model, preds=self.preds, out_path=path_xai)
        for i in range(0, len(self.sampleList)):
            path_xai_file = os.path.join(os.path.join(self.tmp_data.name, "xai"),
                                         self.sampleList[i])
            self.assertTrue(os.path.exists(path_xai_file))
            img = image_loader(sample=self.sampleList[i],
                               path_imagedir=self.tmp_data.name,
                               image_format=self.datagen.image_format)
            hm = image_loader(sample=self.sampleList[i],
                              path_imagedir=os.path.join(self.tmp_data.name, "xai"),
                              image_format=self.datagen.image_format)
            self.assertTrue(np.array_equal(img.shape, hm.shape))
            self.assertFalse(np.array_equal(img, hm))

    def test_Decoder_allclasses_output(self):
        imgs, hms = xai_decoder(self.datagen, self.model, preds=None,
                                out_path=None)
        self.assertTrue(np.array_equal(imgs.shape, (10, 32, 32, 3)))
        self.assertTrue(np.array_equal(hms.shape, (10, 4, 32, 32)))

    def test_Decoder_allclasses_visualize(self):
        path_xai = os.path.join(self.tmp_data.name, "xai")
        xai_decoder(self.datagen, self.model, preds=None, out_path=path_xai)
        for i in range(0, len(self.sampleList)):
            sample = self.sampleList[i]
            for c in range(0, 4):
                xai_file = sample[:-4] + ".class_" + str(c) + sample[-4:]
                img = image_loader(sample=self.sampleList[i],
                                   path_imagedir=self.tmp_data.name,
                                   image_format=self.datagen.image_format)
                hm = image_loader(sample=xai_file,
                                  path_imagedir=os.path.join(self.tmp_data.name, "xai"),
                                  image_format=self.datagen.image_format)
                self.assertTrue(np.array_equal(img.shape, hm.shape))
                self.assertFalse(np.array_equal(img, hm))

    def test_Decoder_directoryInterface(self):
        # Create imaging data with subdirectories
        tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                               suffix=".data")
        for i in range(0, 5):
            os.mkdir(os.path.join(tmp_data.name, "class_" + str(i)))
        # Fill subdirectories with images
        for i in range(0, 25):
            img = np.random.rand(32, 32, 3) * 255
            img_pillow = Image.fromarray(img.astype(np.uint8))
            index = "image.sample_" + str(i) + ".png"
            label_dir = "class_" + str((i % 5))
            path_sample = os.path.join(tmp_data.name, label_dir, index)
            img_pillow.save(path_sample)
        ds = input_interface(interface="directory", path_imagedir=tmp_data.name)
        (index_list, _, nclasses, _, _) = ds
        # Create Data Generator
        datagen = DataGenerator(index_list,  tmp_data.name,
                                labels=None, resize=None,
                                grayscale=False, batch_size=3)
        # Create Neural Network model
        model = NeuralNetwork(n_labels=nclasses, channels=3, input_shape=(32,32),
                              architecture="2D.Vanilla")
        path_xai = os.path.join(tmp_data.name, "xai_directory")
        xai_decoder(datagen, model, preds=None, out_path=path_xai)

    #-------------------------------------------------#
    #            XAI Visualization: Heatmap           #
    #-------------------------------------------------#
    def test_Visualizer(self):
        image = self.image[0]
        heatmap = np.random.rand(32, 32)
        path_xai = os.path.join(self.tmp_data.name, "xai_test.png")
        visualize_heatmap(image, heatmap, out_path=path_xai, alpha=0.4)
        self.assertTrue(os.path.exists(path_xai))
        img = image_loader(sample=self.sampleList[0],
                           path_imagedir=self.tmp_data.name,
                           image_format=self.datagen.image_format)
        hm = image_loader(sample="xai_test.png",
                          path_imagedir=self.tmp_data.name,
                          image_format=None)
        self.assertTrue(np.array_equal(img.shape, hm.shape))
        self.assertFalse(np.array_equal(img, hm))

    #-------------------------------------------------#
    #              XAI Methods: Grad-Cam              #
    #-------------------------------------------------#
    def test_XAImethod_GradCam_init(self):
        GradCAM(self.model.model)
        xai_dict["gradcam"](self.model.model)
        xai_dict["gc"](self.model.model)

    def test_XAImethod_GradCam_heatmap(self):
        xai_method = GradCAM(self.model.model)
        for i in range(4):
            hm = xai_method.compute_heatmap(image=self.image, class_index=i)
            self.assertTrue(np.array_equal(hm.shape, (2,2)))

    def test_XAImethod_GradCam_decoder(self):
        imgs, hms = xai_decoder(self.datagen, self.model, method="gradcam")
        self.assertTrue(np.array_equal(hms.shape, (10, 4, 32, 32)))

    #-------------------------------------------------#
    #             XAI Methods: Grad-Cam++             #
    #-------------------------------------------------#
    def test_XAImethod_GradCamPP_init(self):
        GradCAMpp(self.model.model)
        xai_dict["gradcam++"](self.model.model)
        xai_dict["gc++"](self.model.model)

    def test_XAImethod_GradCamPP_heatmap(self):
        xai_method = GradCAMpp(self.model.model)
        for i in range(4):
            hm = xai_method.compute_heatmap(image=self.image, class_index=i)
            self.assertTrue(np.array_equal(hm.shape, (2,2)))

    def test_XAImethod_GradCamPP_decoder(self):
        imgs, hms = xai_decoder(self.datagen, self.model, method="gradcam++")
        self.assertTrue(np.array_equal(hms.shape, (10, 4, 32, 32)))

    #-------------------------------------------------#
    #            XAI Methods: Saliency Maps           #
    #-------------------------------------------------#
    def test_XAImethod_SaliencyMap_init(self):
        SaliencyMap(self.model.model)
        xai_dict["saliency"](self.model.model)
        xai_dict["sm"](self.model.model)

    def test_XAImethod_SaliencyMap_heatmap(self):
        xai_method = SaliencyMap(self.model.model)
        for i in range(4):
            hm = xai_method.compute_heatmap(image=self.image, class_index=i)
            self.assertTrue(np.array_equal(hm.shape, (32,32)))

    def test_XAImethod_SaliencyMap_decoder(self):
        imgs, hms = xai_decoder(self.datagen, self.model, method="saliency")
        self.assertTrue(np.array_equal(hms.shape, (10, 4, 32, 32)))

    #-------------------------------------------------#
    #       XAI Methods: Guided Backpropagation       #
    #-------------------------------------------------#
    def test_XAImethod_GuidedBackprop_init(self):
        GuidedBackpropagation(self.model.model)
        xai_dict["guidedbackprop"](self.model.model)
        xai_dict["gb"](self.model.model)

    def test_XAImethod_GuidedBackprop_heatmap(self):
        xai_method = GuidedBackpropagation(self.model.model)
        for i in range(4):
            hm = xai_method.compute_heatmap(image=self.image, class_index=i)
            self.assertTrue(np.array_equal(hm.shape, (32,32)))

    def test_XAImethod_GuidedBackprop_decoder(self):
        imgs, hms = xai_decoder(self.datagen, self.model, method="guidedbackprop")
        self.assertTrue(np.array_equal(hms.shape, (10, 4, 32, 32)))

    #-------------------------------------------------#
    #        XAI Methods: Integrated Gradients        #
    #-------------------------------------------------#
    def test_XAImethod_IntegratedGradients_init(self):
        IntegratedGradients(self.model.model)
        xai_dict["IntegratedGradients"](self.model.model)
        xai_dict["ig"](self.model.model)

    def test_XAImethod_IntegratedGradients_heatmap(self):
        xai_method = IntegratedGradients(self.model.model)
        for i in range(4):
            hm = xai_method.compute_heatmap(image=self.image, class_index=i)
            self.assertTrue(np.array_equal(hm.shape, (32,32)))

    def test_XAImethod_IntegratedGradients_decoder(self):
        imgs, hms = xai_decoder(self.datagen, self.model, method="IntegratedGradients")
        self.assertTrue(np.array_equal(hms.shape, (10, 4, 32, 32)))

    #-------------------------------------------------#
    #           XAI Methods: Guided Grad-CAM          #
    #-------------------------------------------------#
    def test_XAImethod_GuidedGradCAM_init(self):
        GuidedGradCAM(self.model.model)
        xai_dict["GuidedGradCAM"](self.model.model)
        xai_dict["ggc"](self.model.model)

    def test_XAImethod_GuidedGradCAM_heatmap(self):
        xai_method = GuidedGradCAM(self.model.model)
        for i in range(4):
            hm = xai_method.compute_heatmap(image=self.image, class_index=i)
            self.assertTrue(np.array_equal(hm.shape, (32,32)))

    def test_XAImethod_GuidedGradCAM_decoder(self):
        imgs, hms = xai_decoder(self.datagen, self.model, method="GuidedGradCAM")
        self.assertTrue(np.array_equal(hms.shape, (10, 4, 32, 32)))

    #-------------------------------------------------#
    #        XAI Methods: Occlusion Sensitivity       #
    #-------------------------------------------------#
    def test_XAImethod_OcclusionSensitivity_init(self):
        OcclusionSensitivity(self.model.model, patch_size=16)
        xai_dict["OcclusionSensitivity"](self.model.model)
        xai_dict["os"](self.model.model)

    def test_XAImethod_OcclusionSensitivity_heatmap(self):
        xai_method = OcclusionSensitivity(self.model.model, patch_size=16)
        for i in range(4):
            hm = xai_method.compute_heatmap(image=self.image, class_index=i)
            self.assertTrue(np.array_equal(hm.shape, (32,32)))

    def test_XAImethod_OcclusionSensitivity_decoder(self):
        imgs, hms = xai_decoder(self.datagen, self.model, method="OcclusionSensitivity")
        self.assertTrue(np.array_equal(hms.shape, (10, 4, 32, 32)))

    #-------------------------------------------------#
    #              XAI Methods: LIME Con              #
    #-------------------------------------------------#
    def test_XAImethod_LimeCon_init(self):
        LimeCon(self.model.model, num_samples=10)
        xai_dict["LimeCon"](self.model.model)
        xai_dict["lc"](self.model.model)

    def test_XAImethod_LimeCon_heatmap(self):
        xai_method = LimeCon(self.model.model, num_samples=10)
        for i in range(4):
            hm = xai_method.compute_heatmap(image=self.image, class_index=i)
            self.assertTrue(np.array_equal(hm.shape, (32,32)))

    def test_XAImethod_LimeCon_decoder(self):
        xai_method = LimeCon(self.model.model, num_samples=10)
        imgs, hms = xai_decoder(self.datagen, self.model, method=xai_method)
        self.assertTrue(np.array_equal(hms.shape, (10, 4, 32, 32)))

    #-------------------------------------------------#
    #              XAI Methods: LIME Pro              #
    #-------------------------------------------------#
    def test_XAImethod_LimePro_init(self):
        LimePro(self.model.model, num_samples=10)
        xai_dict["LimePro"](self.model.model)
        xai_dict["lp"](self.model.model)

    def test_XAImethod_LimePro_heatmap(self):
        xai_method = LimePro(self.model.model, num_samples=10)
        for i in range(4):
            hm = xai_method.compute_heatmap(image=self.image, class_index=i)
            self.assertTrue(np.array_equal(hm.shape, (32,32)))

    def test_XAImethod_LimePro_decoder(self):
        xai_method = LimePro(self.model.model, num_samples=10)
        imgs, hms = xai_decoder(self.datagen, self.model, method=xai_method)
        self.assertTrue(np.array_equal(hms.shape, (10, 4, 32, 32)))



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
# Abstract Base Class for Architectures
from aucmedi.neural_network.architectures.arch_base import Architecture_Base

#-----------------------------------------------------#
#                    Architectures                    #
#-----------------------------------------------------#
# Vanilla Classifier
from aucmedi.neural_network.architectures.volume.vanilla import Vanilla
# DenseNet
from aucmedi.neural_network.architectures.volume.densenet121 import DenseNet121
from aucmedi.neural_network.architectures.volume.densenet169 import DenseNet169
from aucmedi.neural_network.architectures.volume.densenet201 import DenseNet201
# ResNet
from aucmedi.neural_network.architectures.volume.resnet18 import ResNet18
from aucmedi.neural_network.architectures.volume.resnet34 import ResNet34
from aucmedi.neural_network.architectures.volume.resnet50 import ResNet50
from aucmedi.neural_network.architectures.volume.resnet101 import ResNet101
from aucmedi.neural_network.architectures.volume.resnet152 import ResNet152
# ResNeXt
from aucmedi.neural_network.architectures.volume.resnext50 import ResNeXt50
from aucmedi.neural_network.architectures.volume.resnext101 import ResNeXt101
# MobileNet
from aucmedi.neural_network.architectures.volume.mobilenet import MobileNet
from aucmedi.neural_network.architectures.volume.mobilenetv2 import MobileNetV2
# VGG
from aucmedi.neural_network.architectures.volume.vgg16 import VGG16
from aucmedi.neural_network.architectures.volume.vgg19 import VGG19
# ConvNeXt
from aucmedi.neural_network.architectures.volume.convnext_tiny import ConvNeXtTiny
from aucmedi.neural_network.architectures.volume.convnext_small import ConvNeXtSmall
from aucmedi.neural_network.architectures.volume.convnext_base import ConvNeXtBase
from aucmedi.neural_network.architectures.volume.convnext_large import ConvNeXtLarge

#-----------------------------------------------------#
#       Access Functions to Architecture Classes      #
#-----------------------------------------------------#
# Architecture Dictionary
architecture_dict = {
    "Vanilla": Vanilla,
    "DenseNet121": DenseNet121,
    "DenseNet169": DenseNet169,
    "DenseNet201": DenseNet201,
    "ResNet18": ResNet18,
    "ResNet34": ResNet34,
    "ResNet50": ResNet50,
    "ResNet101": ResNet101,
    "ResNet152": ResNet152,
    "ResNeXt50": ResNeXt50,
    "ResNeXt101": ResNeXt101,
    "MobileNet": MobileNet,
    "MobileNetV2": MobileNetV2,
    "VGG16": VGG16,
    "VGG19": VGG19,
    "ConvNeXtTiny": ConvNeXtTiny,
    "ConvNeXtSmall": ConvNeXtSmall,
    "ConvNeXtBase": ConvNeXtBase,
    "ConvNeXtLarge": ConvNeXtLarge,
}
""" Dictionary of implemented 3D Architectures Methods in AUCMEDI.

    The base key (str) or an initialized Architecture can be passed to the [NeuralNetwork][aucmedi.neural_network.model.NeuralNetwork] class as `architecture` parameter.

    ???+ example "Example"
        ```python title="Recommended via NeuralNetwork class"
        my_model = NeuralNetwork(n_labels=4, channels=1, architecture="3D.ResNet50",
                                  input_shape(128,128,128), activation_output="softmax")
        ```

        ```python title="Manual via architecture_dict import"
        from aucmedi.neural_network.architectures import Classifier, architecture_dict

        classification_head = Classifier(n_labels=4, activation_output="softmax")
        my_arch = architecture_dict["3D.ResNet50"](classification_head,
                                                   channels=1, input_shape=(128,128,128))
        my_model = NeuralNetwork(n_labels=None, channels=None, architecture=my_arch)
        ```

        ```python title="Manual via module import"
        from aucmedi.neural_network.architectures import Classifier
        from aucmedi.neural_network.architectures.volume import ResNet50

        classification_head = Classifier(n_labels=4, activation_output="softmax")
        my_arch = ResNet50(classification_head,
                                        channels=1, input_shape=(128,128,128))

        my_model = NeuralNetwork(n_labels=None, channels=None, architecture=my_arch)
        ```

    ???+ warning
        If passing an architecture key to the NeuralNetwork class, be aware that you have to add "3D." in front of it.

        For example:
        ```python
        # for the volume architecture "ResNeXt101"
        architecture="3D.ResNeXt101"
        ```

    Architectures are based on the abstract base class [aucmedi.neural_network.architectures.arch_base.Architecture_Base][].
"""

# List of implemented architectures
architectures = list(architecture_dict.keys())

#-----------------------------------------------------#
#       Meta Information of Architecture Classes      #
#-----------------------------------------------------#
# Utilized standardize mode of architectures required for Transfer Learning
supported_standardize_mode = {
    "Vanilla": "z-score",
    "DenseNet121": "torch",
    "DenseNet169": "torch",
    "DenseNet201": "torch",
    "ResNet18": "grayscale",
    "ResNet34": "grayscale",
    "ResNet50": "grayscale",
    "ResNet101": "grayscale",
    "ResNet152": "grayscale",
    "ResNeXt50": "grayscale",
    "ResNeXt101": "grayscale",
    "MobileNet": "tf",
    "MobileNetV2": "tf",
    "VGG16": "caffe",
    "VGG19": "caffe",
    "ConvNeXtTiny": None,
    "ConvNeXtSmall": None,
    "ConvNeXtBase": None,
    "ConvNeXtLarge": None,
}
""" Dictionary of recommended [Standardize][aucmedi.data_processing.subfunctions.standardize] techniques for 3D Architectures Methods in AUCMEDI.

    The base key (str) can be passed to the [DataGenerator][aucmedi.data_processing.data_generator.DataGenerator] as `standardize_mode` parameter.

    ???+ info
        If training a new model from scratch, any Standardize technique can be used at will. <br>
        However, if training via transfer learning, it is required to use the recommended Standardize technique!

    ???+ example "Example"
        ```python title="Recommended via the NeuralNetwork class"
        my_model = NeuralNetwork(n_labels=8, channels=3, architecture="3D.DenseNet121")

        my_dg = DataGenerator(samples, "images_dir/", labels=None,
                              resize=my_model.meta_input,                  # (64, 64, 64)
                              standardize_mode=my_model.meta_standardize)  # "torch"
        ```

        ```python title="Manual via supported_standardize_mode import"
        from aucmedi.neural_network.architectures import supported_standardize_mode
        sf_norm = supported_standardize_mode["3D.DenseNet121"]

        my_dg = DataGenerator(samples, "images_dir/", labels=None,
                              resize=(64, 64, 64),                         # (64, 64, 64)
                              standardize_mode=sf_norm)                    # "torch"
        ```

    ???+ warning
        If using an architecture key for the supported_standardize_mode dictionary, be aware that you have to add "3D." in front of it.

        For example:
        ```python
        # for the volume architecture "ResNeXt101"
        from aucmedi.neural_network.architectures import supported_standardize_mode
        sf_norm = supported_standardize_mode["3D.ResNeXt101"]
        ```
"""



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
#External libraries
import unittest
import tempfile
import os
from PIL import Image
import numpy as np
from unittest.mock import patch
import sys
from shutil import which
#Internal libraries
from aucmedi.automl.main import main
from aucmedi.automl.cli import *
from aucmedi.automl import parse_cli

#-----------------------------------------------------#
#                Unittest: AutoML CLI                 #
#-----------------------------------------------------#
class AutoML_CLI(unittest.TestCase):
    # Create random imaging and classification data
    @classmethod
    def setUpClass(self):
        np.random.seed(1234)
        # Create temporary directory-based imaging data
        self.tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                                    suffix=".data")
        for i in range(0, 5):
            os.mkdir(os.path.join(self.tmp_data.name, "class_" + str(i)))
        # Fill subdirectories with images
        for i in range(0, 25):
            img = np.random.rand(16, 16, 3) * 255
            img_pillow = Image.fromarray(img.astype(np.uint8))
            index = "image.sample_" + str(i) + ".png"
            label_dir = "class_" + str((i % 5))
            path_sample = os.path.join(self.tmp_data.name, label_dir, index)
            img_pillow.save(path_sample)
        # Create temporary model directory
        self.tmp_model = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                                     suffix=".model")

    #-------------------------------------------------#
    #                  CLI Hub: Core                  #
    #-------------------------------------------------#
    def test_core_empty(self):
        if which("aucmedi") is None : return    # only check unittesting for build (install via pip)
        args = ["aucmedi"]
        with patch.object(sys, "argv", args):
            with self.assertRaises(SystemExit) as se:
                main()
            self.assertEqual(se.exception.code, 1)

    def test_core_version(self):
        if which("aucmedi") is None : return    # only check unittesting for build (install via pip)
        with patch.object(sys, "argv", ["aucmedi", "-v"]):
            with self.assertRaises(SystemExit) as se:
                main()
        with patch.object(sys, "argv", ["aucmedi", "--version"]):
            with self.assertRaises(SystemExit) as se:
                main()

    def test_core_help(self):
        if which("aucmedi") is None : return    # only check unittesting for build (install via pip)
        with patch.object(sys, "argv", ["aucmedi", "-h"]):
            with self.assertRaises(SystemExit) as se:
                main()
        with patch.object(sys, "argv", ["aucmedi", "--help"]):
            with self.assertRaises(SystemExit) as se:
                main()

    #-------------------------------------------------#
    #                CLI Hub: Training                #
    #-------------------------------------------------#
    def test_training_functionality(self):
        if which("aucmedi") is None : return    # only check unittesting for build (install via pip)
        args = ["aucmedi", "training"]
        args_config = ["--path_imagedir", self.tmp_data.name,
                       "--epochs", "1",
                       "--architecture", "Vanilla",
                       "--path_modeldir", self.tmp_model.name]
        with patch.object(sys, "argv", args + args_config):
            main()

    def test_training_args(self):
        if which("aucmedi") is None : return    # only check unittesting for build (install via pip)
        args = ["aucmedi", "training"]
        args_config = ["--path_imagedir", self.tmp_data.name]
        # Build and run CLI functions
        with patch.object(sys, "argv", args + args_config):
            parser, subparsers = cli_core()
            cli_training(subparsers)
            args = parser.parse_args()
            config_cli = parse_cli(args)
        # Define possible config parameters
        config_map = ["path_imagedir",
                      "path_gt",
                      "path_modeldir",
                      "analysis",
                      "ohe",
                      "three_dim",
                      "shape_3D",
                      "epochs",
                      "batch_size",
                      "workers",
                      "metalearner",
                      "architecture",
                     ]
        # Check existence
        for c in config_map:
            self.assertTrue(c in config_cli)

    #-------------------------------------------------#
    #               CLI Hub: Prediction               #
    #-------------------------------------------------#
    def test_prediction_functionality(self):
        if which("aucmedi") is None : return    # only check unittesting for build (install via pip)
        # Training
        args = ["aucmedi", "training"]
        args_config = ["--path_imagedir", self.tmp_data.name,
                       "--epochs", "1",
                       "--architecture", "Vanilla",
                       "--path_modeldir", self.tmp_model.name]
        with patch.object(sys, "argv", args + args_config):
            main()
        # Prediction
        tmp_output = tempfile.NamedTemporaryFile(mode="w",
                                                 prefix="tmp.aucmedi.",
                                                 suffix=".pred.csv")
        args = ["aucmedi", "prediction"]
        args_config = ["--path_imagedir", os.path.join(self.tmp_data.name,
                                                       "class_0"),
                       "--path_modeldir", self.tmp_model.name,
                       "--path_pred", tmp_output.name]
        with patch.object(sys, "argv", args + args_config):
            main()

    def test_prediction_args(self):
        if which("aucmedi") is None : return    # only check unittesting for build (install via pip)
        args = ["aucmedi", "prediction"]
        args_config = ["--path_imagedir", os.path.join(self.tmp_data.name,
                                                         "class_0")]
        # Build and run CLI functions
        with patch.object(sys, "argv", args + args_config):
            parser, subparsers = cli_core()
            cli_prediction(subparsers)
            args = parser.parse_args()
            config_cli = parse_cli(args)
        # Define possible config parameters
        config_map = ["path_imagedir",
                      "path_modeldir",
                      "path_pred",
                      "xai_method",
                      "xai_directory",
                      "batch_size",
                      "workers",
                     ]
        # Check existence
        for c in config_map:
            self.assertTrue(c in config_cli)

    #-------------------------------------------------#
    #               CLI Hub: Evaluation               #
    #-------------------------------------------------#
    def test_evaluation_functionality(self):
        if which("aucmedi") is None : return    # only check unittesting for build (install via pip)
        # Training
        args = ["aucmedi", "training"]
        args_config = ["--path_imagedir", self.tmp_data.name,
                       "--epochs", "1",
                       "--architecture", "Vanilla",
                       "--path_modeldir", self.tmp_model.name]
        with patch.object(sys, "argv", args + args_config):
            main()
        # Prediction
        tmp_output = tempfile.NamedTemporaryFile(mode="w",
                                                 prefix="tmp.aucmedi.",
                                                 suffix=".pred.csv")
        args = ["aucmedi", "prediction"]
        args_config = ["--path_imagedir", os.path.join(self.tmp_data.name,
                                                       "class_0"),
                       "--path_modeldir", self.tmp_model.name,
                       "--path_pred", tmp_output.name]
        with patch.object(sys, "argv", args + args_config):
            main()
        # Evaluation
        tmp_eval = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                               suffix=".eval")
        args = ["aucmedi", "evaluation"]
        args_config = ["--path_imagedir", self.tmp_data.name,
                       "--path_pred", tmp_output.name,
                       "--path_evaldir", tmp_eval.name]
        with patch.object(sys, "argv", args + args_config):
            main()

    def test_evaluation_args(self):
        if which("aucmedi") is None : return    # only check unittesting for build (install via pip)
        args = ["aucmedi", "evaluation"]
        args_config = ["--path_imagedir", self.tmp_data.name]
        # Build and run CLI functions
        with patch.object(sys, "argv", args + args_config):
            parser, subparsers = cli_core()
            cli_evaluation(subparsers)
            args = parser.parse_args()
            config_cli = parse_cli(args)
        # Define possible config parameters
        config_map = ["path_gt",
                      "path_pred",
                      "path_evaldir",
                      "ohe",
                     ]
        # Check existence
        for c in config_map:
            self.assertTrue(c in config_cli)



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
from tensorflow.keras.models import load_model
from tensorflow.keras.optimizers import Adam
import numpy as np
# Internal libraries/scripts
from aucmedi.neural_network.architectures import architecture_dict, \
                                                 supported_standardize_mode, \
                                                 Classifier

#-----------------------------------------------------#
#            Neural Network (model) class             #
#-----------------------------------------------------#
# Class which represents the Neural Network
class NeuralNetwork:
    """ Neural Network class providing functionality for handling all model methods.

    This class is the third of the three pillars of AUCMEDI.

    ??? info "Pillars of AUCMEDI"
        - [aucmedi.data_processing.io_data.input_interface][]
        - [aucmedi.data_processing.data_generator.DataGenerator][]
        - [aucmedi.neural_network.model.NeuralNetwork][]

    With an initialized Neural Network model instance, it is possible to run training and predictions.

    ??? example "Example: How to use"
        ```python
        # Initialize model
        model = NeuralNetwork(n_labels=8, channels=3, architecture="2D.ResNet50")
        # Do some training
        datagen_train = DataGenerator(samples[:100], "images_dir/", labels=class_ohe[:100],
                                      resize=model.meta_input, standardize_mode=model.meta_standardize)
        model.train(datagen_train, epochs=50)
        # Do some predictions
        datagen_test = DataGenerator(samples[100:150], "images_dir/", labels=None,
                                     resize=model.meta_input, standardize_mode=model.meta_standardize)
        preds = model.predict(datagen_test)
        ```

    ??? example "Example: How to select an Architecture"
        ```python
        # 2D architecture
        my_model_a = NeuralNetwork(n_labels=8, channels=3, architecture="2D.DenseNet121")
        # 3D architecture for multi-label classification (sigmoid activation)
        my_model_b = NeuralNetwork(n_labels=8, channels=3, architecture="3D.ResNet50",
                                    activation_output="sigmoid")
        # 2D architecture with custom input_shape
        my_model_c = NeuralNetwork(n_labels=8, channels=3, architecture="2D.Xception",
                                    input_shape=(512,512))
        ```

    ??? note "List of implemented Architectures"
        AUCMEDI provides a large library of state-of-the-art and ready-to-use architectures.

        - 2D Architectures: [aucmedi.neural_network.architectures.image][]
        - 3D Architectures: [aucmedi.neural_network.architectures.volume][]

    ??? note "Classification Types"
        | Type                       | Activation Function                                             |
        | -------------------------- | --------------------------------------------------------------- |
        | Binary classification      | `activation_output="softmax"`: Only a single class is correct.  |
        | Multi-class classification | `activation_output="softmax"`: Only a single class is correct.  |
        | Multi-label classification | `activation_output="sigmoid"`: Multiple classes can be correct. |

        Defined by the [Classifier][aucmedi.neural_network.architectures.classifier] of an
        [Architecture][aucmedi.neural_network.architectures].

    ??? example "Example: How to obtain required parameters for the DataGenerator?"
        Be aware that the input_size and standardize_mode are just recommendations and
        can be changed by desire. <br>
        However, the recommended parameter are required for transfer learning.

        ```python title="Recommended way"
        my_model = NeuralNetwork(n_labels=8, channels=3, architecture="2D.DenseNet121")

        my_dg = DataGenerator(samples, "images_dir/", labels=None,
                              resize=my_model.meta_input,                  # (224,224)
                              standardize_mode=my_model.meta_standardize)  # "torch"
        ```

        ```python title="Manual way"
        from aucmedi.neural_network.architectures import Classifier, \
                                                         architecture_dict, \
                                                         supported_standardize_mode

        classification_head = Classifier(n_labels=4, activation_output="softmax")
        my_arch = architecture_dict["3D.DenseNet121"](classification_head,
                                                      channels=1,
                                                      input_shape=(128,128,128))

        my_model = NeuralNetwork(n_labels=None, channels=None, architecture=my_arch)

        from aucmedi.neural_network.architectures import supported_standardize_mode
        sf_norm = supported_standardize_mode["3D.DenseNet121"]
        my_dg = DataGenerator(samples, "images_dir/", labels=None,
                              resize=(128,128,128),                        # (128,128,128)
                              standardize_mode=sf_norm)                    # "torch"
        ```

    ??? example "Example: How to integrate metadata in AUCMEDI?"
        ```python
        from aucmedi import *
        import numpy as np

        my_metadata = np.random.rand(len(samples), 10)

        my_model = NeuralNetwork(n_labels=8, channels=3, architecture="2D.DenseNet121",
                                  meta_variables=10)

        my_dg = DataGenerator(samples, "images_dir/",
                              labels=None, metadata=my_metadata,
                              resize=my_model.meta_input,                  # (224,224)
                              standardize_mode=my_model.meta_standardize)  # "torch"
        ```
    """
    def __init__(self, n_labels, channels, input_shape=None, architecture=None,
                 pretrained_weights=False, loss="categorical_crossentropy",
                 metrics=["categorical_accuracy"], activation_output="softmax",
                 fcl_dropout=True, meta_variables=None, learning_rate=0.0001,
                 verbose=1):
        """ Initialization function for creating a Neural Network (model) object.

        Args:
            n_labels (int):                         Number of classes/labels (important for the last layer).
            channels (int):                         Number of channels. Grayscale:1 or RGB:3.
            input_shape (tuple):                    Input shape of the batch imaging data (including channel axis).
                                                    If None is provided, the default input_shape for the architecture is selected
                                                    from the architecture dictionary.
            architecture (str or Architecture):     Key (str) or instance of a neural network model Architecture class instance.
                                                    If a string is provided, the corresponding architecture is selected from the architecture dictionary.
                                                    A string has to begin with either '3D.' or '2D' depending on the classification task.
                                                    By default, a 2D Vanilla Model is used as architecture.
            pretrained_weights (bool):              Option whether to utilize pretrained weights e.g. from ImageNet.
            loss (Metric Function):                 The metric function which is used as loss for training.
                                                    Any Metric Function defined in Keras, in aucmedi.neural_network.loss_functions or any custom
                                                    metric function, which follows the Keras metric guidelines, can be used.
            metrics (list of Metric Functions):     List of one or multiple Metric Functions, which will be shown during training.
                                                    Any Metric Function defined in Keras or any custom metric function, which follows the Keras
                                                    metric guidelines, can be used.
            activation_output (str):                Activation function which should be used in the classification head
                                                    ([Classifier][aucmedi.neural_network.architectures.classifier]).
                                                    Based on [https://www.tensorflow.org/api_docs/python/tf/keras/activations](https://www.tensorflow.org/api_docs/python/tf/keras/activations).
            fcl_dropout (bool):                     Option whether to utilize an additional Dense & Dropout layer in the classification head
                                                    ([Classifier][aucmedi.neural_network.architectures.classifier]).
            meta_variables (int):                   Number of metadata variables, which should be included in the classification head.
                                                    If `None`is provided, no metadata integration block will be added to the classification head
                                                    ([Classifier][aucmedi.neural_network.architectures.classifier]).
            learning_rate (float):                  Learning rate in which weights of the neural network will be updated.
            verbose (int):                          Option (0/1) how much information should be written to stdout.

        ???+ danger
            Class attributes can be modified also after initialization, at will.
            However, be aware of unexpected adverse effects (experimental)!

        Attributes:
            tf_epochs (int, default=5):             Transfer Learning configuration: Number of epochs with frozen layers except classification head.
            tf_lr_start (float, default=1e-4):      Transfer Learning configuration: Starting learning rate for frozen layer fitting.
            tf_lr_end (float, default=1e-5):        Transfer Learning configuration: Starting learning rate after layer unfreezing.
            meta_input (tuple of int):              Meta variable: Input shape of architecture which can be passed to a DataGenerator. For example: (224, 224).
            meta_standardize (str):                 Meta variable: Recommended standardize_mode of architecture which can be passed to a DataGenerator.
                                                    For example: "torch".
        """
        # Cache parameters
        self.n_labels = n_labels
        self.channels = channels
        self.loss = loss
        self.metrics = metrics
        self.learning_rate = learning_rate
        self.pretrained_weights = pretrained_weights
        self.activation_output = activation_output
        self.fcl_dropout = fcl_dropout
        self.meta_variables = meta_variables
        self.verbose = verbose

        # Assemble architecture parameters
        arch_paras = {"channels":channels,
                      "pretrained_weights": pretrained_weights}
        if input_shape is not None : arch_paras["input_shape"] = input_shape
        # Assemble classifier parameters
        classifier_paras = {"n_labels": n_labels, "fcl_dropout": fcl_dropout,
                            "activation_output": activation_output}
        if meta_variables is not None:
            classifier_paras["meta_variables"] = meta_variables
        # Initialize classifier for the classification head
        arch_paras["classification_head"] = Classifier(**classifier_paras)
        # Initialize architecture if None provided
        if architecture is None:
            self.architecture = architecture_dict["2D.Vanilla"](**arch_paras)
            self.meta_standardize = "z-score"
        # Initialize passed architecture from aucmedi library
        elif isinstance(architecture, str) and architecture in architecture_dict:
            self.architecture = architecture_dict[architecture](**arch_paras)
            self.meta_standardize = supported_standardize_mode[architecture]
        # Initialize passed architecture as parameter
        else:
            self.architecture = architecture
            self.meta_standardize = None

        # Build model utilizing the selected architecture
        self.model = self.architecture.create_model()

        # Compile model
        self.model.compile(optimizer=Adam(learning_rate=learning_rate),
                           loss=self.loss, metrics=self.metrics)

        # Obtain final input shape
        self.input_shape = self.architecture.input          # e.g. (224, 224, 3)
        self.meta_input = self.architecture.input[:-1]      # e.g. (224, 224) -> for DataGenerator
        # Cache starting weights
        self.initialization_weights = self.model.get_weights()

    #---------------------------------------------#
    #               Class Variables               #
    #---------------------------------------------#
    # Transfer Learning configurations
    tf_epochs = 10
    tf_lr_start = 1e-4
    tf_lr_end = 1e-5

    #---------------------------------------------#
    #                  Training                   #
    #---------------------------------------------#
    # Training the Neural Network model
    def train(self, training_generator, validation_generator=None, epochs=20,
              iterations=None, callbacks=[], class_weights=None,
              transfer_learning=False):
        """ Fitting function for the Neural Network model performing a training process.

        It is also possible to pass custom Callback classes in order to obtain more information.

        If an optional validation [DataGenerator][aucmedi.data_processing.data_generator.DataGenerator]
        is provided, a validation set is analyzed regularly during the training process (after each epoch).

        The transfer learning training runs two fitting processes.
        The first one with frozen base model layers and a high learning rate,
        whereas the second one with unfrozen layers and a small learning rate.

        ??? info "Keras History Objects for Transfer Learning"
            For the transfer learning training, two Keras history objects will be created.

            However, in order to provide consistency with the single training without transfer learning,
            only a single history dictionary will be returned.

            For differentiation prefixes are added in front of the corresponding logging keys:
            ```
            - History Start ->  prefix : tl     for "transfer learning"
            - History End   ->  prefix : ft     for "fine tuning"
            ```

        Args:
            training_generator (DataGenerator):     A data generator which will be used for training.
            validation_generator (DataGenerator):   A data generator which will be used for validation.
            epochs (int):                           Number of epochs. A single epoch is defined as one iteration through
                                                    the complete data set.
            iterations (int):                       Number of iterations (batches) in a single epoch.
            callbacks (list of Callback classes):   A list of Callback classes for custom evaluation.
            class_weights (dictionary or list):     A list or dictionary of float values to handle class unbalance.
            transfer_learning (bool):               Option whether a transfer learning training should be performed. If true, a minimum of 5 epochs will be trained.

        Returns:
            history (dict):                   A history dictionary from a Keras history object which contains several logs.
        """
        # Adjust number of iterations in training DataGenerator to allow repitition
        if iterations is not None : training_generator.set_length(iterations)
        # Running a standard training process
        if not transfer_learning:
            # Run training process with the Keras fit function
            history = self.model.fit(training_generator,
                                     validation_data=validation_generator,
                                     callbacks=callbacks, epochs=epochs,
                                     steps_per_epoch=iterations,
                                     class_weight=class_weights,
                                     verbose=self.verbose)
            # Return logged history object
            history_out = history.history
        # Running a transfer learning training process
        else:
            # Freeze all base model layers (all layers after "avg_pool")
            lever = False
            for layer in reversed(self.model.layers):
                if not lever and layer.name == "avg_pool" : lever = True
                elif lever : layer.trainable = False
            # Compile model with high learning rate
            self.model.compile(optimizer=Adam(learning_rate=self.tf_lr_start),
                               loss=self.loss, metrics=self.metrics)
            # Run first training with frozen layers
            history_start = self.model.fit(training_generator,
                                           validation_data=validation_generator,
                                           callbacks=callbacks,
                                           epochs=self.tf_epochs,
                                           steps_per_epoch=iterations,
                                           class_weight=class_weights,
                                           verbose=self.verbose)
            # Unfreeze base model layers again
            for layer in self.model.layers:
                layer.trainable = True
            # Compile model with lower learning rate
            self.model.compile(optimizer=Adam(learning_rate=self.tf_lr_end),
                               loss=self.loss, metrics=self.metrics)
            # Run second training with unfrozed layers
            history_end = self.model.fit(training_generator,
                                         validation_data=validation_generator,
                                         callbacks=callbacks, epochs=epochs,
                                         initial_epoch=self.tf_epochs,
                                         steps_per_epoch=iterations,
                                         class_weight=class_weights,
                                         verbose=self.verbose)
            # Combine logged history objects
            hs = {"tl_" + k: v for k, v in history_start.history.items()}       # prefix : tl for transfer learning
            he = {"ft_" + k: v for k, v in history_end.history.items()}         # prefix : ft for fine tuning
            history = {**hs, **he}
            # Return combined history objects
            history_out = history
        # Reset number of iterations of the training DataGenerator
        if iterations is not None : training_generator.reset_length()
        # Return fitting history
        return history_out

    #---------------------------------------------#
    #                 Prediction                  #
    #---------------------------------------------#
    def predict(self, prediction_generator):
        """ Prediction function for the Neural Network model.

        The fitted model will predict classifications for the provided [DataGenerator][aucmedi.data_processing.data_generator.DataGenerator].

        Args:
            prediction_generator (DataGenerator):   A data generator which will be used for inference.

        Returns:
            preds (numpy.ndarray):                  A NumPy array of predictions formatted with shape (n_samples, n_labels).
        """
        # Run inference process with the Keras predict function
        preds = self.model.predict(prediction_generator,
                                   verbose=self.verbose)
        # Output predictions results
        return preds

    #---------------------------------------------#
    #               Model Management              #
    #---------------------------------------------#
    # Re-initialize model weights
    def reset_weights(self):
        """ Re-initialize weights of the neural network model.

        Useful for training multiple models with the same NeuralNetwork object.
        """
        self.model.set_weights(self.initialization_weights)

    # Dump model to file
    def dump(self, file_path):
        """ Store model to disk.

        Recommended to utilize the file format ".keras".

        Args:
            file_path (str):    Path to store the model on disk.
        """
        self.model.save(file_path)

    # Load model from file
    def load(self, file_path, custom_objects={}):
        """ Load neural network model and its weights from a file.

        After loading, the model will be compiled.

        If loading a model in ".keras" format, it is not necessary to define any custom_objects.

        Args:
            file_path (str):            Input path, from which the model will be loaded.
            custom_objects (dict):      Dictionary of custom objects for compiling
                                        (e.g. non-TensorFlow based loss functions or architectures).
        """
        # Create model input path
        self.model = load_model(file_path, custom_objects, compile=False)
        # Compile model
        self.model.compile(optimizer=Adam(learning_rate=self.learning_rate),
                           loss=self.loss, metrics=self.metrics)



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                    Documentation                    #
#-----------------------------------------------------#
""" Library of implemented Subfunctions in AUCMEDI.

    A Subfunction is a preprocessing method, which is automatically applied on all samples
    if provided to a [DataGenerator][aucmedi.data_processing.data_generator.DataGenerator].

    Image preprocessing is defined as a method or technique which modify the image before passing it to the neural network model.
    The aim of preprocessing methods is to extensively increase performance due to simplification of information.
    Common preprocessing methods range from intensity value normalization to image resizing.

    ???+ info
        The DataGenerator applies the list of Subfunctions **sequentially** on the data set.

    The Subfunctions [Resize][aucmedi.data_processing.subfunctions.resize] and
    [Standardize][aucmedi.data_processing.subfunctions.standardize] are integrated into the
    DataGenerator base due to its requirement in any medical image classification pipeline.

    ???+ example "Example"
        ```python
        # Import Subfunctions
        from aucmedi.data_processing.subfunctions import *

        # Select desired Subfunctions
        sf_crop = Crop(shape=(128, 164), mode="center")
        sf_padding = Padding(mode="square")
        # Pack them into a list
        sf_list = [sf_crop, sf_padding]                 # Subfunctions will be applied in provided list order

        # Pass the list to the DataGenerator
        train_gen = DataGenerator(samples=index_list,
                                  path_imagedir="my_images/",
                                  labels=class_ohe,
                                  resize=(512,512),                # Call the integrated resize Subfunction
                                  standardize_mode="grayscale",    # Call the integrated standardize Subfunction
                                  subfunctions=sf_list)            # Pass desired Subfunctions
        ```

    Subfunctions are based on the abstract base class [Subfunction_Base][aucmedi.data_processing.subfunctions.sf_base.Subfunction_Base],
    which allow simple integration of custom preprocessing methods.
"""
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
from aucmedi.data_processing.subfunctions.standardize import Standardize
from aucmedi.data_processing.subfunctions.resize import Resize
from aucmedi.data_processing.subfunctions.padding import Padding
from aucmedi.data_processing.subfunctions.crop import Crop
from aucmedi.data_processing.subfunctions.color_constancy import ColorConstancy
from aucmedi.data_processing.subfunctions.clip import Clip
from aucmedi.data_processing.subfunctions.chromer import Chromer



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
import numpy as np
# Internal libraries/scripts
from aucmedi.ensemble.aggregate.agg_base import Aggregate_Base

#-----------------------------------------------------#
#               Aggregate: Majority Vote              #
#-----------------------------------------------------#
class MajorityVote(Aggregate_Base):
    """ Aggregate function based on majority vote.

    This class should be passed to an ensemble function/class for combining predictions.
    """
    #---------------------------------------------#
    #                Initialization               #
    #---------------------------------------------#
    def __init__(self):
        # No hyperparameter adjustment required for this method, therefore skip
        pass

    #---------------------------------------------#
    #                  Aggregate                  #
    #---------------------------------------------#
    def aggregate(self, preds):
        # Count votes
        votes = np.argmax(preds, axis=1)
        # Identify majority
        majority_vote = np.argmax(np.bincount(votes))
        # Create prediction based on majority vote
        pred = np.zeros((preds.shape[1]))
        pred[majority_vote] = 1
        # Return prediction
        return pred



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
import os
# Internal libraries
import aucmedi.data_processing.io_interfaces as io

#-----------------------------------------------------#
#                   Static Variables                  #
#-----------------------------------------------------#
ACCEPTABLE_IMAGE_FORMATS = ["jpeg", "jpg", "tif", "tiff", "png", "bmp", "gif",
                            "npy", "nii", "gz", "mha"]
""" List of accepted image formats. """

#-----------------------------------------------------#
#             Input Interface for AUCMEDI             #
#-----------------------------------------------------#
def input_interface(interface, path_imagedir, path_data=None, training=True,
                    ohe=False, image_format=None, **kwargs):
    """ Data Input Interface for all automatically extract various information of dataset structures.

    Different image file structures and annotation information are processed by
    corresponding format interfaces. These extracted information can be parsed to the
    [DataGenerator][aucmedi.data_processing.data_generator.DataGenerator] and the
    [NeuralNetwork][aucmedi.neural_network.model.NeuralNetwork].

    The input_interface() function is the first of the three pillars of AUCMEDI.

    ??? info "Pillars of AUCMEDI"
        - [aucmedi.data_processing.io_data.input_interface][]
        - [aucmedi.data_processing.data_generator.DataGenerator][]
        - [aucmedi.neural_network.model.NeuralNetwork][]

    Basically a wrapper function for calling the correct format interface,
    which loads a dataset from disk via the associated format parser.

    Possible format interfaces: `["csv", "json", "directory"]`

    ???+ info "Format Interfaces"
        | Interface      | Internal Function                                                    | Description                                  |
        | -------------- | -------------------------------------------------------------------- | -------------------------------------------- |
        |  `"csv"`       | [io_csv()][aucmedi.data_processing.io_interfaces.io_csv]             | Storing class annotations in a CSV file.     |
        |  `"directory"` | [io_directory()][aucmedi.data_processing.io_interfaces.io_directory] | Storing class annotations in subdirectories. |
        |  `"json"`      | [io_json()][aucmedi.data_processing.io_interfaces.io_json]           | Storing class annotations in a JSON file.    |

    ???+ example
        ```python
        # AUCMEDI library
        from aucmedi import *

        # Initialize input data reader
        ds = input_interface(interface="csv",                       # Interface type
                             path_imagedir="dataset/images/",
                             path_data="dataset/annotations.csv",
                             ohe=False, col_sample="ID", col_class="diagnosis")
        (index_list, class_ohe, nclasses, class_names, image_format) = ds

        # Pass variables to other AUCMEDI pillars like DataGenerator
        datagen = DataGenerator(samples=index_list,                 # from input_interface()
                                path_imagedir="dataset/images/",
                                labels=class_ohe,                   # from input_interface()
                                image_format=image_format)          # from input_interface()
        ```

    Args:
        path_imagedir (str):            Path to the directory containing the images.
        interface (str):                String defining format interface for loading/storing data.
        path_data (str):                Path to the index/class annotation file if required. (csv/json)
        training (bool):                Boolean option whether annotation data is available.
        ohe (bool):                     Boolean option whether annotation data is sparse categorical or one-hot encoded.
        image_format (str):             Force to use a specific image format. By default, image format is determined automatically.
        **kwargs (dict):                Additional parameters for the format interfaces.

    Returns:
        index_list (list of str):       List of sample/index encoded as Strings. Required in DataGenerator as `samples`.
        class_ohe (numpy.ndarray):      Classification list as One-Hot encoding. Required in DataGenerator as `labels`.
        class_n (int):                  Number of classes. Required in NeuralNetwork for Architecture design as `n_labels`.
        class_names (list of str):      List of names for corresponding classes. Used for later prediction storage or evaluation.
        image_format (str):             Image format to add at the end of the sample index for image loading. Required in DataGenerator.
    """
    # Transform selected interface to lower case
    interface = interface.lower()
    # Pass image format if provided
    if image_format != None : allowed_image_formats = [image_format]
    else : allowed_image_formats = ACCEPTABLE_IMAGE_FORMATS
    # Verify if provided interface is valid
    if interface not in ["csv", "json", "directory"]:
        raise Exception("Unknown interface code provided.", interface)
    # Verify that annotation file is available if CSV/JSON interface is used
    if interface in ["csv", "json"] and path_data is None:
        raise Exception("No annotation file provided for CSV/JSON interface!")

    # Initialize parameter dictionary
    parameters = {"path_data": path_data,
                  "path_imagedir": path_imagedir,
                  "allowed_image_formats": allowed_image_formats,
                  "training": training, "ohe": ohe}
    # Identify correct dataset loader and parameters for CSV format
    if interface == "csv":
        ds_loader = io.csv_loader
        additional_parameters = ["ohe_range", "col_sample", "col_class"]
        for para in additional_parameters:
            if para in kwargs : parameters[para] = kwargs[para]
    # Identify correct dataset loader and parameters for JSON format
    elif interface == "json" : ds_loader = io.json_loader
    # Identify correct dataset loader and parameters for directory format
    elif interface == "directory":
        ds_loader = io.directory_loader
        del parameters["ohe"]
        del parameters["path_data"]

    # Load the dataset with the selected format interface and return results
    return ds_loader(**parameters)



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External Libraries
import numpy as np
import tensorflow as tf
# Internal Libraries
from aucmedi.xai.methods.xai_base import XAImethod_Base

#-----------------------------------------------------#
#     Gradient-weighted Class Activation Mapping      #
#-----------------------------------------------------#
class GradCAM(XAImethod_Base):
    """ XAI Method for Gradient-weighted Class Activation Mapping (Grad-CAM).

    Normally, this class is used internally in the [aucmedi.xai.decoder.xai_decoder][] in the AUCMEDI XAI module.

    ??? abstract "Reference - Implementation #1"
        Author: François Chollet <br>
        Date: April 26, 2020 <br>
        [https://keras.io/examples/vision/grad_cam/](https://keras.io/examples/vision/grad_cam/) <br>

    ??? abstract "Reference - Implementation #2"
        Author: Adrian Rosebrock <br>
        Date: March 9, 2020 <br>
        [https://www.pyimagesearch.com/2020/03/09/grad-cam-visualize-class-activation-maps-with-keras-tensorflow-and-deep-learning/](https://www.pyimagesearch.com/2020/03/09/grad-cam-visualize-class-activation-maps-with-keras-tensorflow-and-deep-learning/) <br>

    ??? abstract "Reference - Publication"
        Ramprasaath R. Selvaraju, Michael Cogswell, Abhishek Das, Ramakrishna Vedantam, Devi Parikh, Dhruv Batra. 7 Oct 2016.
        Grad-CAM: Visual Explanations from Deep Networks via Gradient-based Localization.
        <br>
        [https://arxiv.org/abs/1610.02391](https://arxiv.org/abs/1610.02391)

    This class provides functionality for running the compute_heatmap function,
    which computes a Grad-CAM heatmap for an image with a model.
    """
    def __init__(self, model, layerName=None):
        """ Initialization function for creating a Grad-CAM as XAI Method object.

        Args:
            model (keras.model):               Keras model object.
            layerName (str):                   Layer name of the convolutional layer for heatmap computation.
        """
        # Cache class parameters
        self.model = model
        self.layerName = layerName
        # Try to find output layer if not defined
        if self.layerName is None : self.layerName = self.find_output_layer()

    #---------------------------------------------#
    #            Identify Output Layer            #
    #---------------------------------------------#
    def find_output_layer(self):
        """ Internal function. Applied if `layerName==None`.

        Identify last/final convolutional layer in neural network architecture.
        This layer is used to obtain activation outputs / feature map.
        """
        # Iterate over all layers
        for layer in reversed(self.model.layers):
            # Check to see if the layer has a 4D output -> Return layer
            if len(layer.output.shape) == 4:
                return layer.name
        # Otherwise, throw exception
        raise ValueError("Could not find 4D layer. Cannot apply Grad-CAM.")

    #---------------------------------------------#
    #             Heatmap Computation             #
    #---------------------------------------------#
    def compute_heatmap(self, image, class_index, eps=1e-8):
        """ Core function for computing the Grad-CAM heatmap for a provided image and for specific classification outcome.

        ???+ attention
            Be aware that the image has to be provided in batch format.

        Args:
            image (numpy.ndarray):              Image matrix encoded as NumPy Array (provided as one-element batch).
            class_index (int):                  Classification index for which the heatmap should be computed.
            eps (float):                        Epsilon for rounding.

        The returned heatmap is encoded within a range of [0,1]

        ???+ attention
            The shape of the returned heatmap is 2D -> batch and channel axis will be removed.

        Returns:
            heatmap (numpy.ndarray):            Computed Grad-CAM for provided image.
        """
        # Gradient model construction
        gradModel = tf.keras.models.Model(inputs=[self.model.inputs],
                         outputs=[self.model.get_layer(self.layerName).output,
                                  self.model.output])
        # Compute gradient for desierd class index
        with tf.GradientTape() as tape:
            inputs = tf.cast(image, tf.float32)
            (conv_out, preds) = gradModel(inputs)
            loss = preds[:, class_index]
        grads = tape.gradient(loss, conv_out)
        # Averaged output gradient based on feature map of last conv layer
        pooled_grads = tf.reduce_mean(grads, axis=(0, 1, 2))
        # Normalize gradients via "importance"
        heatmap = conv_out[0] @ pooled_grads[..., tf.newaxis]
        heatmap = tf.squeeze(heatmap).numpy()

        # Intensity normalization to [0,1]
        numer = heatmap - np.min(heatmap)
        denom = (heatmap.max() - heatmap.min()) + eps
        heatmap = numer / denom

        # Return the resulting heatmap
        return heatmap



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External Libraries
import numpy as np
import tensorflow as tf
# Internal Libraries
from aucmedi.xai.methods.xai_base import XAImethod_Base

#-----------------------------------------------------#
#                XAI Method: Grad-Cam++               #
#-----------------------------------------------------#
class GradCAMpp(XAImethod_Base):
    """ XAI Method for Grad-CAM++.

    Normally, this class is used internally in the [aucmedi.xai.decoder.xai_decoder][] in the AUCMEDI XAI module.

    ??? abstract "Reference - Implementation"
        Author: Samson Woof <br>
        GitHub Profile: [https://github.com/samson6460](https://github.com/samson6460) <br>
        Date: May 21, 2020 <br>
        [https://github.com/samson6460/tf_keras_gradcamplusplus](https://github.com/samson6460/tf_keras_gradcamplusplus) <br>

    ??? abstract "Reference - Publication"
        Aditya Chattopadhay; Anirban Sarkar; Prantik Howlader; Vineeth N Balasubramanian. 07 May 2018.
        Grad-CAM++: Generalized Gradient-Based Visual Explanations for Deep Convolutional Networks.
        <br>
        [https://ieeexplore.ieee.org/document/8354201](https://ieeexplore.ieee.org/document/8354201)

    This class provides functionality for running the compute_heatmap function,
    which computes a Grad-CAM++ heatmap for an image with a model.
    """
    def __init__(self, model, layerName=None):
        """ Initialization function for creating a Grad-CAM++ as XAI Method object.

        Args:
            model (keras.model):               Keras model object.
            layerName (str):                   Layer name of the convolutional layer for heatmap computation.
        """
        # Cache class parameters
        self.model = model
        self.layerName = layerName
        # Try to find output layer if not defined
        if self.layerName is None : self.layerName = self.find_output_layer()

    #---------------------------------------------#
    #            Identify Output Layer            #
    #---------------------------------------------#
    def find_output_layer(self):
        """ Internal function. Applied if `layerName==None`.

        Identify last/final convolutional layer in neural network architecture.
        This layer is used to obtain activation outputs / feature map.
        """
        # Iterate over all layers
        for layer in reversed(self.model.layers):
            # Check to see if the layer has a 4D output -> Return layer
            if len(layer.output.shape) == 4:
                return layer.name
        # Otherwise, throw exception
        raise ValueError("Could not find 4D layer. Cannot apply Grad-CAM++.")

    #---------------------------------------------#
    #             Heatmap Computation             #
    #---------------------------------------------#
    def compute_heatmap(self, image, class_index, eps=1e-8):
        """ Core function for computing the Grad-CAM++ heatmap for a provided image and for specific classification outcome.

        ???+ attention
            Be aware that the image has to be provided in batch format.

        Args:
            image (numpy.ndarray):              Image matrix encoded as NumPy Array (provided as one-element batch).
            class_index (int):                  Classification index for which the heatmap should be computed.
            eps (float):                        Epsilon for rounding.

        The returned heatmap is encoded within a range of [0,1]

        ???+ attention
            The shape of the returned heatmap is 2D -> batch and channel axis will be removed.

        Returns:
            heatmap (numpy.ndarray):            Computed Grad-CAM++ for provided image.
        """
        # Gradient model construction
        gradModel = tf.keras.models.Model(inputs=[self.model.inputs],
                         outputs=[self.model.get_layer(self.layerName).output,
                                  self.model.output])

        # Compute gradient for desierd class index
        with tf.GradientTape() as gtape1:
            with tf.GradientTape() as gtape2:
                with tf.GradientTape() as gtape3:
                    inputs = tf.cast(image, tf.float32)
                    (conv_output, preds) = gradModel(inputs)
                    output = preds[:, class_index]
                    conv_first_grad = gtape3.gradient(output, conv_output)
                conv_second_grad = gtape2.gradient(conv_first_grad, conv_output)
            conv_third_grad = gtape1.gradient(conv_second_grad, conv_output)
        global_sum = np.sum(conv_output, axis=(0, 1, 2))

        # Normalize constants
        alpha_num = conv_second_grad[0]
        alpha_denom = conv_second_grad[0]*2.0 + conv_third_grad[0]*global_sum
        alpha_denom = np.where(alpha_denom != 0.0, alpha_denom, eps)
        alphas = alpha_num / alpha_denom
        alpha_normalization_constant = np.sum(alphas, axis=(0,1))
        alphas /= alpha_normalization_constant

        # Deep Linearization weighting
        weights = np.maximum(conv_first_grad[0], 0.0)
        deep_linearization_weights = np.sum(weights*alphas, axis=(0,1))
        heatmap = np.sum(deep_linearization_weights*conv_output[0], axis=2)

        # Intensity normalization to [0,1]
        numer = heatmap - np.min(heatmap)
        denom = (heatmap.max() - heatmap.min()) + eps
        heatmap = numer / denom

        # Return the resulting heatmap
        return heatmap



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External Libraries
import numpy as np
import pandas as pd
import os
from plotnine import *
# Internal libraries/scripts
from aucmedi.evaluation.metrics import *

#-----------------------------------------------------#
#            Evaluation - Plot Performance            #
#-----------------------------------------------------#
def evaluate_performance(preds,
                         labels,
                         out_path,
                         show=False,
                         class_names=None,
                         multi_label=False,
                         metrics_threshold=0.5,
                         suffix=None,
                         store_csv=True,
                         plot_barplot=True,
                         plot_confusion_matrix=True,
                         plot_roc_curve=True):
    """ Function for automatic performance evaluation based on model predictions.

    ???+ example
        ```python
        # Import libraries
        from aucmedi import *
        from aucmedi.evaluation import *

        # Load data
        ds = input_interface(interface="csv",                       # Interface type
                             path_imagedir="dataset/images/",
                             path_data="dataset/annotations.csv",
                             ohe=False, col_sample="ID", col_class="diagnosis")
        (samples, class_ohe, nclasses, class_names, image_format) = ds

        # Initialize model
        model = NeuralNetwork(n_labels=8, channels=3, architecture="2D.ResNet50")

        # Do some predictions
        datagen_test = DataGenerator(samples, "dataset/images/", labels=None,
                                     resize=model.meta_input, standardize_mode=model.meta_standardize)
        preds = model.predict(datagen_test)

        # Pass predictions to evaluation function
        evaluate_performance(preds, class_ohe, out_path="./", class_names=class_names)
        ```

    Created files in directory of `out_path`:

    - with `store_csv`: "metrics.performance.csv"
    - with `plot_barplot`: "plot.performance.barplot.png"
    - with `plot_confusion_matrix`: "plot.performance.confusion_matrix.png"
    - with `plot_roc_curve`: "plot.performance.roc.png"

    ???+ info "Preview for Bar Plot"
        ![Evaluation_Performance_Barplot](../../images/evaluation.plot.performance.barplot.png)

        Predictions based on [ISIC 2019 Challenge](https://challenge.isic-archive.com/landing/2019/)
        utilizing a DenseNet121.

    ???+ info "Preview for Confusion Matrix"
        ![Evaluation_Performance_ConfusionMatrix](../../images/evaluation.plot.performance.confusion_matrix.png)

        Predictions based on [ISIC 2019 Challenge](https://challenge.isic-archive.com/landing/2019/)
        utilizing a DenseNet121.

    ???+ info "Preview for ROC Curve"
        ![Evaluation_Performance_ROCcurve](../../images/evaluation.plot.performance.roc.png)

        Predictions based on [ISIC 2019 Challenge](https://challenge.isic-archive.com/landing/2019/)
        utilizing a DenseNet121.

    Args:
        preds (numpy.ndarray):          A NumPy array of predictions formatted with shape (n_samples, n_labels). Provided by
                                        [NeuralNetwork][aucmedi.neural_network.model].
        labels (numpy.ndarray):         Classification list with One-Hot Encoding. Provided by
                                        [input_interface][aucmedi.data_processing.io_data.input_interface].
        out_path (str):                 Path to directory in which plotted figures are stored.
        show (bool):                    Option, whether to also display the generated charts.
        class_names (list of str):      List of names for corresponding classes. Used for evaluation. Provided by
                                        [input_interface][aucmedi.data_processing.io_data.input_interface].
                                        If not provided (`None` provided), class indices will be used.
        multi_label (bool):             Option, whether task is multi-label based (has impact on evaluation).
        metrics_threshold (float):      Only required if 'multi_label==True`. Threshold value if prediction is positive.
                                        Used in metric computation for CSV and bar plot.
        suffix (str):                   Special suffix to add in the created figure filename.
        store_csv (bool):               Option, whether to generate a CSV file containing various metrics.
        plot_barplot (bool):            Option, whether to generate a bar plot of various metrics.
        plot_confusion_matrix (bool):   Option, whether to generate a confusion matrix plot.
        plot_roc_curve (bool):          Option, whether to generate a ROC curve plot.

    Returns:
        metrics (pandas.DataFrame):     Dataframe containing all computed metrics (except ROC).
    """
    # Identify number of labels
    n_labels = labels.shape[-1]
    # Identify prediction threshold
    if multi_label : threshold = metrics_threshold
    else : threshold = None

    # Compute metrics
    metrics = compute_metrics(preds, labels, n_labels, threshold)
    cm = compute_confusion_matrix(preds, labels, n_labels)
    fpr_list, tpr_list = compute_roc(preds, labels, n_labels)

    # Rename columns in metrics dataframe
    class_mapping = {}
    if class_names is not None:
        for c in range(len(class_names)):
            class_mapping[c] = class_names[c]
        metrics["class"].replace(class_mapping, inplace=True)
    if class_names is None : metrics["class"] = pd.Categorical(metrics["class"])

    # Store metrics to CSV
    if store_csv:
        evalby_csv(metrics, out_path, class_names, suffix=suffix)

    # Generate bar plot
    if plot_barplot:
        evalby_barplot(metrics, out_path, class_names, show=show, suffix=suffix)

    # Generate confusion matrix plot
    if plot_confusion_matrix and not multi_label:
        evalby_confusion_matrix(cm, out_path, class_names, show=show, suffix=suffix)

    # Generate ROC curve
    if plot_roc_curve:
        evalby_rocplot(fpr_list, tpr_list, out_path, class_names, show=show, suffix=suffix)

    # Return metrics
    return metrics

#-----------------------------------------------------#
#      Evaluation Performance - Confusion Matrix      #
#-----------------------------------------------------#
def evalby_confusion_matrix(confusion_matrix, out_path, class_names,
                            show=False,
                            suffix=None):

    # Convert confusion matrix to a Pandas dataframe
    rawcm = pd.DataFrame(confusion_matrix)
    # Tidy dataframe
    if class_names is None or len(class_names) != confusion_matrix.shape[0]:
        class_names = list(range(0, confusion_matrix.shape[0]))
    rawcm.index = class_names
    rawcm.columns = class_names

    # Preprocess dataframe
    dt = rawcm.div(rawcm.sum(axis=1), axis=0).fillna(0) * 100
    dt = dt.round(decimals=2)
    dt.reset_index(drop=False, inplace=True)
    dt = dt.melt(id_vars=["index"], var_name="pd", value_name="score")
    dt.rename(columns={"index": "gt"}, inplace=True)

    # Generate confusion matrix
    fig = (ggplot(dt, aes("pd", "gt", fill="score"))
                  + geom_tile(color="white", size=1.5)
                  + geom_text(aes("pd", "gt", label="score"), color="black")
                  + ggtitle("Performance Evaluation: Confusion Matrix")
                  + xlab("Prediction")
                  + ylab("Ground Truth")
                  + scale_fill_gradient(low="white", high="royalblue",
                                        limits=[0, 100])
                  + guides(fill = guide_colourbar(title="%",
                                                  barwidth=10,
                                                  barheight=50))
                  + theme_bw()
                  + theme(axis_text_x = element_text(angle = 45, vjust = 1,
                                                     hjust = 1)))

    # Store figure to disk
    filename = "plot.performance.confusion_matrix"
    if suffix is not None : filename += "." + str(suffix)
    filename += ".png"
    fig.save(filename=filename, path=out_path, width=10, height=9, dpi=200)

    # Plot figure
    if show : print(fig)

#-----------------------------------------------------#
#          Evaluation Performance - Barplots          #
#-----------------------------------------------------#
def evalby_barplot(metrics, out_path, class_names, show=False, suffix=None):
    # Remove confusion matrix from metric dataframe
    df_metrics = metrics[~metrics["metric"].isin(["TN", "FN", "FP", "TP"])]
    df_metrics["class"] = pd.Categorical(df_metrics["class"])

    # Generate metric results
    fig = (ggplot(df_metrics, aes("class", "score", fill="class"))
              + geom_col(stat='identity', width=0.6, color="black",
                         position = position_dodge(width=0.6))
              + ggtitle("Performance Evaluation: Metric Overview")
              + facet_wrap("metric")
              + coord_flip()
              + xlab("")
              + ylab("Score")
              + scale_y_continuous(limits=[0, 1], breaks=np.arange(0, 1.1, 0.1))
              + scale_fill_discrete(name="Classes")
              + theme_bw())

    # Store figure to disk
    filename = "plot.performance.barplot"
    if suffix is not None : filename += "." + str(suffix)
    filename += ".png"
    fig.save(filename=filename, path=out_path, width=12, height=9, dpi=200)

    # Plot figure
    if show : print(fig)

#-----------------------------------------------------#
#          Evaluation Performance - ROC plot          #
#-----------------------------------------------------#
def evalby_rocplot(fpr_list, tpr_list, out_path, class_names, show=False, suffix=None):
    # Initialize result dataframe
    df_roc = pd.DataFrame(data=[fpr_list, tpr_list], dtype=object)
    # Preprocess dataframe
    df_roc = df_roc.transpose()
    df_roc = df_roc.apply(pd.Series.explode)
    # Rename columns
    class_mapping = {}
    if class_names is not None:
        for c in range(len(class_names)):
            class_mapping[c] = class_names[c]
        df_roc.rename(index=class_mapping, inplace=True)
    df_roc = df_roc.reset_index()
    df_roc.rename(columns={"index": "class", 0: "FPR", 1: "TPR"}, inplace=True)
    df_roc["class"] = pd.Categorical(df_roc["class"])
    # Convert from object to float
    df_roc["FPR"] = df_roc["FPR"].astype(float)
    df_roc["TPR"] = df_roc["TPR"].astype(float)

    # Generate roc results
    fig = (ggplot(df_roc, aes("FPR", "TPR", color="class"))
               + geom_line(size=1.0)
               + geom_abline(intercept=0, slope=1, color="black",
                             linetype="dashed")
               + ggtitle("Performance Evaluation: ROC Curves")
               + xlab("False Positive Rate")
               + ylab("True Positive Rate")
               + scale_x_continuous(limits=[0, 1], breaks=np.arange(0,1.1,0.1))
               + scale_y_continuous(limits=[0, 1], breaks=np.arange(0,1.1,0.1))
               + scale_color_discrete(name="Classes")
               + theme_bw())

    # Store figure to disk
    filename = "plot.performance.roc"
    if suffix is not None : filename += "." + str(suffix)
    filename += ".png"
    fig.save(filename=filename, path=out_path, width=10, height=9, dpi=200)

    # Plot figure
    if show : print(fig)

#-----------------------------------------------------#
#          Evaluation Performance - CSV file          #
#-----------------------------------------------------#
def evalby_csv(metrics, out_path, class_names, suffix=None):
    # Obtain filename to
    filename = "metrics.performance"
    if suffix is not None : filename += "." + str(suffix)
    filename += ".csv"
    path_csv = os.path.join(out_path, filename)

    # Store file to disk
    metrics.to_csv(path_csv, index=False)



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                    Documentation                    #
#-----------------------------------------------------#
""" The classification variant of the Vanilla architecture.

No intensive hardware requirements, which makes it ideal for debugging.

| Architecture Variable    | Value                      |
| ------------------------ | -------------------------- |
| Key in architecture_dict | "2D.Vanilla"               |
| Input_shape              | (224, 224)                 |
| Standardization          | "z-score"                  |

???+ abstract "Reference - Implementation"
    [https://github.com/wanghsinwei/isic-2019/](https://github.com/wanghsinwei/isic-2019/) <br>
"""
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
from tensorflow.keras import Input
from tensorflow.keras.layers import Conv2D, MaxPooling2D
# Internal libraries
from aucmedi.neural_network.architectures import Architecture_Base

#-----------------------------------------------------#
#                 Vanilla Architecture                #
#-----------------------------------------------------#
class Vanilla(Architecture_Base):
    #---------------------------------------------#
    #                   __init__                  #
    #---------------------------------------------#
    def __init__(self, classification_head, channels, input_shape=(224, 224),
                 pretrained_weights=False):
        self.classifier = classification_head
        self.input = input_shape + (channels,)
        self.pretrained_weights = pretrained_weights

    #---------------------------------------------#
    #                Create Model                 #
    #---------------------------------------------#
    def create_model(self):
        # Initialize input
        model_input = Input(shape=self.input)

        # Add 4x convolutional layers with increasing filters
        model_base = Conv2D(filters=32, kernel_size=3, padding='same',
                            activation='relu')(model_input)
        model_base = MaxPooling2D(pool_size=2)(model_base)

        model_base = Conv2D(filters=64, kernel_size=3, padding='same',
                            activation='relu')(model_base)
        model_base = MaxPooling2D(pool_size=2)(model_base)

        model_base = Conv2D(filters=128, kernel_size=3, padding='same',
                            activation='relu')(model_base)
        model_base = MaxPooling2D(pool_size=2)(model_base)

        model_base = Conv2D(filters=256, kernel_size=3, padding='same',
                            activation='relu')(model_base)
        model_base = MaxPooling2D(pool_size=2)(model_base)

        # Add classification head
        model = self.classifier.build(model_input=model_input,
                                      model_output=model_base)

        # Return created model
        return model



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
from abc import ABC, abstractmethod

#-----------------------------------------------------#
#         Abstract Base Class for XAI Methods         #
#-----------------------------------------------------#
class XAImethod_Base(ABC):
    """ An abstract base class for a XAI Method class.

    Normally, this class is used internally in the xai_decoder function in the AUCMEDI XAI module.

    This class provides functionality for running the compute_heatmap function,
    which computes a heatmap for an image with a model.

    ???+ example "Create a custom XAImethod"
        ```python
        from aucmedi.xai.methods.xai_base import XAImethod_Base

        class My_custom_XAImethod(XAImethod_Base):
            def __init__(self, model, layerName=None):
                pass

            def compute_heatmap(self, image, class_index, eps=1e-8):
                pass
        ```

    ???+ info "Required Functions"
        | Function            | Description                                |
        | ------------------- | ------------------------------------------ |
        | `__init__()`        | Object creation function.                  |
        | `compute_heatmap()` | Application of the XAI Method on an image. |

    """
    #---------------------------------------------#
    #                Initialization               #
    #---------------------------------------------#
    @abstractmethod
    def __init__(self, model, layerName=None):
        """ Initialization function for creating a XAI Method object.
        ```
        __init__(model, layerName=None)
        ```

        Args:
            model (keras.model):               Keras model object.
            layerName (str):                Layer name of the convolutional layer for heatmap computation.
        """
        pass

    #---------------------------------------------#
    #             Heatmap Computation             #
    #---------------------------------------------#
    def compute_heatmap(self, image, class_index, eps=1e-8):
        """ Core function for computing the XAI heatmap for a provided image and for specific classification outcome.

        ???+ attention
            Be aware that the image has to be provided in batch format.

        Args:
            image (numpy.ndarray):              Image matrix encoded as NumPy Array (provided as one-element batch).
            class_index (int):                  Classification index for which the heatmap should be computed.
            eps (float):                        Epsilon for rounding.

        The returned heatmap should be encoded within a range of [0,1]

        ???+ attention
            The shape of the returned heatmap is 2D -> batch and channel axis will be removed.

        Returns:
            heatmap (numpy.ndarray):            Computed XAI heatmap for provided image.
        """
        pass



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
import os
import numpy as np
import json
from tensorflow.keras.metrics import AUC
from tensorflow.keras.callbacks import ModelCheckpoint, CSVLogger, \
                                       ReduceLROnPlateau, EarlyStopping
# Internal libraries
from aucmedi import *
from aucmedi.data_processing.io_loader import image_loader, sitk_loader
from aucmedi.sampling import sampling_split
from aucmedi.utils.class_weights import *
from aucmedi.data_processing.subfunctions import *
from aucmedi.neural_network.loss_functions import *
from aucmedi.ensemble import *
from aucmedi.evaluation import evaluate_fitting

#-----------------------------------------------------#
#            Building Blocks for Training             #
#-----------------------------------------------------#
def block_train(config):
    """ Internal code block for AutoML training.

    This function is called by the Command-Line-Interface (CLI) of AUCMEDI.

    Args:
        config (dict):                      Configuration dictionary containing all required
                                            parameters for performing an AutoML training.

    The following attributes are stored in the `config` dictionary:

    Attributes:
        path_imagedir (str):                Path to the directory containing the images.
        path_modeldir (str):                Path to the output directory in which fitted models and metadata are stored.
        path_gt (str):                      Path to the index/class annotation file if required. (only for 'csv' interface).
        analysis (str):                     Analysis mode for the AutoML training. Options: `["minimal", "standard", "advanced"]`.
        ohe (bool):                         Boolean option whether annotation data is sparse categorical or one-hot encoded.
        three_dim (bool):                   Boolean, whether data is 2D or 3D.
        shape_3D (tuple of int):            Desired input shape of 3D volume for architecture (will be cropped).
        epochs (int):                       Number of epochs. A single epoch is defined as one iteration through
                                            the complete data set.
        batch_size (int):                   Number of samples inside a single batch.
        workers (int):                      Number of workers/threads which preprocess batches during runtime.
        metalearner (str):                  Key for Metalearner or Aggregate function.
        architecture (str or list of str):  Key (str) of a neural network model Architecture class instance.
    """
    # Obtain interface
    if config["path_gt"] is None : config["interface"] = "directory"
    else : config["interface"] = "csv"
    # Peak into the dataset via the input interface
    ds = input_interface(config["interface"],
                         config["path_imagedir"],
                         path_data=config["path_gt"],
                         training=True,
                         ohe=config["ohe"],
                         image_format=None)
    (index_list, class_ohe, class_n, class_names, image_format) = ds

    # Create output directory
    if not os.path.exists(config["path_modeldir"]):
        os.mkdir(config["path_modeldir"])

    # Identify task (multi-class vs multi-label)
    if np.sum(class_ohe) > class_ohe.shape[0] : config["multi_label"] = True
    else : config["multi_label"] = False

    # Sanity check on multi-label metalearner
    multilabel_metalearner_supported = ["mlp", "k_neighbors", "random_forest",
                                        "weighted_mean", "best_model",
                                        "decision_tree", "mean", "median"]
    if config["multi_label"] and config["analysis"] == "advanced" and \
       config["metalearner"] not in multilabel_metalearner_supported:
        raise ValueError("Non-compatible metalearner selected for multi-label"\
                         + " classification. Supported metalearner:",
                          multilabel_metalearner_supported)

    # Store meta information
    config["class_names"] = class_names
    path_meta = os.path.join(config["path_modeldir"], "meta.training.json")
    with open(path_meta, "w") as json_io:
        json.dump(config, json_io)

    # Define Callbacks
    callbacks = []
    if config["analysis"] == "standard":
        cb_loss = ModelCheckpoint(os.path.join(config["path_modeldir"],
                                               "model.best_loss.keras"),
                                  monitor="val_loss", verbose=1,
                                  save_best_only=True)
        callbacks.append(cb_loss)
    if config["analysis"] in ["minimal", "standard"]:
        cb_cl = CSVLogger(os.path.join(config["path_modeldir"],
                                       "logs.training.csv"),
                          separator=',', append=True)
        callbacks.append(cb_cl)
    if config["analysis"] != "minimal":
        cb_lr = ReduceLROnPlateau(monitor='val_loss', factor=0.1, patience=5,
                              verbose=1, mode='min', min_lr=1e-7)
        cb_es = EarlyStopping(monitor='val_loss', patience=12, verbose=1)
        callbacks.extend([cb_lr, cb_es])

    # Initialize loss function for multi-class
    if not config["multi_label"]:
        # Compute class weights
        class_weights, _ = compute_class_weights(ohe_array=class_ohe)
        # Initialize focal loss
        loss = categorical_focal_loss(class_weights)
    # Initialize loss function for multi-label
    else:
        # Compute class weights
        class_weights = compute_multilabel_weights(ohe_array=class_ohe)
        # Initialize focal loss
        loss = multilabel_focal_loss(class_weights)

    # Define neural network parameters
    nn_paras = {"n_labels": class_n,
                "channels": 3,
                "loss": loss,
                "metrics": [AUC(100)],
                "pretrained_weights": True,
    }
    # Select input shape for 3D
    if config["three_dim"] : nn_paras["input_shape"] = config["shape_3D"]
    # Select task type
    if config["multi_label"] : nn_paras["activation_output"] = "sigmoid"
    else : nn_paras["activation_output"] = "softmax"

    # Initialize Augmentation for 2D image data
    if not config["three_dim"]:
        data_aug = ImageAugmentation(flip=True, rotate=True, scale=False,
                                     brightness=True, contrast=True,
                                     saturation=False, hue=False, crop=False,
                                     grid_distortion=False, compression=False,
                                     gamma=True, gaussian_noise=False,
                                     gaussian_blur=False, downscaling=False,
                                     elastic_transform=True)
    # Initialize Augmentation for 3D volume data
    elif config["three_dim"]:
        data_aug = BatchgeneratorsAugmentation(image_shape=config["shape_3D"],
                        mirror=True, rotate=True, scale=True,
                        elastic_transform=True, gaussian_noise=False,
                        brightness=False, contrast=False, gamma=True)
    else : data_aug = None

    # Subfunctions
    sf_list = []
    if config["three_dim"]:
        sf_norm = Standardize(mode="grayscale")
        sf_pad = Padding(mode="constant", shape=config["shape_3D"])
        sf_crop = Crop(shape=config["shape_3D"], mode="random")
        sf_chromer = Chromer(target="rgb")
        sf_list.extend([sf_norm, sf_pad, sf_crop, sf_chromer])

    # Define parameters for DataGenerator
    paras_datagen = {
        "path_imagedir": config["path_imagedir"],
        "batch_size": config["batch_size"],
        "img_aug": data_aug,
        "subfunctions": sf_list,
        "prepare_images": False,
        "sample_weights": None,
        "seed": None,
        "image_format": image_format,
        "workers": config["workers"],
    }
    if not config["three_dim"] : paras_datagen["loader"] = image_loader
    else : paras_datagen["loader"] = sitk_loader

    # Gather training parameters
    paras_train = {
        "epochs": config["epochs"],
        "iterations": None,
        "callbacks": callbacks,
        "class_weights": None,
        "transfer_learning": True,
    }

    # Apply MIC pipelines
    if config["analysis"] == "minimal":
        # Setup neural network
        if not config["three_dim"] : arch_dim = "2D." + config["architecture"]
        else : arch_dim = "3D." + config["architecture"]
        model = NeuralNetwork(architecture=arch_dim, **nn_paras)

        # Build DataGenerator
        train_gen = DataGenerator(samples=index_list,
                                  labels=class_ohe,
                                  shuffle=True,
                                  resize=model.meta_input,
                                  standardize_mode=model.meta_standardize,
                                  **paras_datagen)

        # Start model training
        hist = model.train(training_generator=train_gen, **paras_train)
        # Store model
        path_model = os.path.join(config["path_modeldir"], "model.last.keras")
        model.dump(path_model)
    elif config["analysis"] == "standard":
        # Setup neural network
        if not config["three_dim"] : arch_dim = "2D." + config["architecture"]
        else : arch_dim = "3D." + config["architecture"]
        model = NeuralNetwork(architecture=arch_dim, **nn_paras)

        # Apply percentage split sampling
        ps_sampling = sampling_split(index_list, class_ohe,
                                     sampling=[0.85, 0.15],
                                     stratified=True, iterative=True,
                                     seed=0)

        # Build DataGenerator
        train_gen = DataGenerator(samples=ps_sampling[0][0],
                                  labels=ps_sampling[0][1],
                                  shuffle=True,
                                  resize=model.meta_input,
                                  standardize_mode=model.meta_standardize,
                                  **paras_datagen)
        val_gen = DataGenerator(samples=ps_sampling[1][0],
                                labels=ps_sampling[1][1],
                                shuffle=False,
                                resize=model.meta_input,
                                standardize_mode=model.meta_standardize,
                                **paras_datagen)

        # Start model training
        hist = model.train(training_generator=train_gen,
                           validation_generator=val_gen,
                           **paras_train)
        # Store model
        path_model = os.path.join(config["path_modeldir"], "model.last.keras")
        model.dump(path_model)
    else:
        # Sanity check of architecutre config
        if not isinstance(config["architecture"], list):
            raise ValueError("key 'architecture' in config has to be a list " \
                             + "if 'advanced' was selected as analysis.")
        # Build multi-model list
        model_list = []
        for arch in config["architecture"]:
            if not config["three_dim"] : arch_dim = "2D." + arch
            else : arch_dim = "3D." + arch
            model_part = NeuralNetwork(architecture=arch_dim, **nn_paras)
            model_list.append(model_part)
        el = Composite(model_list, metalearner=config["metalearner"],
                       k_fold=len(config["architecture"]))

        # Build DataGenerator
        train_gen = DataGenerator(samples=index_list,
                                  labels=class_ohe,
                                  shuffle=True,
                                  resize=None,
                                  standardize_mode=None,
                                  **paras_datagen)
        # Start model training
        hist = el.train(training_generator=train_gen, **paras_train)
        # Store model directory
        el.dump(config["path_modeldir"])

    # Plot fitting history
    evaluate_fitting(train_history=hist, out_path=config["path_modeldir"])



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External Libraries
import numpy as np
import pandas as pd
import os
from plotnine import *

#-----------------------------------------------------#
#              Evaluation - Plot Fitting              #
#-----------------------------------------------------#
def evaluate_fitting(train_history,
                     out_path,
                     monitor=["loss"],
                     prefix_split=".",
                     suffix=None,
                     show=False
                     ):
    """ Function for automatic plot generation providing a training history dictionary.

    !!! info "Preview"
        ![Evaluation_Fitting](../../images/evaluation.plot.fitting_course.png)

    Created filename in directory of `out_path`:

    - without suffix "plot.fitting_course.png"
    - with suffix "plot.fitting_course.SUFFIX.png"

    ???+ example
        ```python
        # Initialize and train a model
        model = NeuralNetwork(n_labels=8, channels=3, architecture="2D.ResNet50")
        history = model.train(datagen_train, datagen_validation, epochs=100)

        # Pass history dict to evaluation function
        evaluate_fitting(history, out_path="./")

        # Figure will be created at: "./plot.fitting_course.png"
        ```

    Args:
        train_history (dict):       A history dictionary from a Keras history object which contains several logs.
        out_path (str):             Path to directory in which plotted figure is stored.
        monitor (list of str):      List of metrics which should be visualized in the fitting plot.
        prefix_split (str):         Split prefix for keys in the history dictionary. Used for Bagging and Stacking.
        suffix (str):               Special suffix to add in the created figure filename.
        show (bool):                Option, whether to also display the generated chart.
    """
    # Convert to pandas dataframe
    hist_prepared = dict([ (k,pd.Series(v)) for k,v in train_history.items() ])
    dt = pd.DataFrame.from_dict(hist_prepared, orient="columns")

    # Identify all selected columns
    selected_cols = []
    for key in train_history:
        for m in monitor:
            if m in key:
                selected_cols.append(key)
                break

    # Add epoch column
    dt["epoch"] = dt.index + 1
    # Melt dataframe
    dt_melted = dt.melt(id_vars=["epoch"],
                        value_vars=selected_cols,
                        var_name="metric",
                        value_name="score")

    # Handle special prefix tags (if split-able by '.')
    if prefix_split is not None:
        for c in selected_cols:
            valid_split = True
            if prefix_split not in c:
                valid_split = False
                break
        if valid_split:
            dt_melted[["prefix", "metric"]] = dt_melted["metric"].str.split(".",
                                                        expand=True)

    # Remove NaN tags
    dt_melted = dt_melted.dropna(axis=0)

    # Preprocess transfer learning tag
    if dt_melted["metric"].str.startswith("tl_").any():
        # filter prefix groups
        filter = dt_melted[dt_melted["metric"].str.startswith("ft_")]

        # if prefix available -> add epochs for each prefix group
        if valid_split:
            # identify number of epochs for each prefix
            filter_tl = dt_melted[dt_melted["metric"].str.startswith("tl_")]
            tl_epochs = filter_tl.groupby(["prefix"])["epoch"].max()
            # compute fine tune epoch update
            group_repeats = filter.groupby(["prefix"]).size()
            if group_repeats.empty : ft_update = 0
            else : ft_update = tl_epochs.repeat(group_repeats).to_numpy()
        # if no prefix available -> add epochs to all ft phases
        else:
            # identify number of epochs global
            filter_tl = dt_melted[dt_melted["metric"].str.startswith("tl_")]
            tl_epochs = filter_tl["epoch"].max()

            # compute fine tune epoch update
            ft_update = tl_epochs

        # update number of epochs
        dt_melted.loc[dt_melted["metric"].str.startswith("ft_"), "epoch"] =\
            filter["epoch"].to_numpy() + ft_update
    else : tl_epochs = None

    # Remove preprocessed transfer learning tag from metric column
    dt_melted["metric"] = dt_melted["metric"].apply(remove_tag, tag="tl_")
    dt_melted["metric"] = dt_melted["metric"].apply(remove_tag, tag="ft_")

    # Preprocess validation tag
    dt_melted["subset"] = np.where(dt_melted["metric"].str.startswith("val_"),
                                   "validation", "training")
    dt_melted["metric"] = dt_melted["metric"].apply(remove_tag, tag="val_")

    # Plot results via plotnine
    fig = (ggplot(dt_melted, aes("epoch", "score", color="subset"))
               + geom_line(size=1)
               + ggtitle("Fitting Curve during Training Process")
               + xlab("Epoch")
               + ylab("Score")
               + scale_colour_discrete(name="Subset")
               + theme_bw()
               + theme(subplots_adjust={'wspace':0.2}))

    if prefix_split is not None and valid_split:
        fig += facet_grid("prefix ~ metric")
    else : fig += facet_wrap("metric", scales="free_y")

    if tl_epochs is not None and valid_split:
        tle_df = tl_epochs.to_frame().reset_index()
        fig += geom_vline(tle_df, aes(xintercept="epoch"))
    elif tl_epochs is not None and not valid_split:
        fig += geom_vline(xintercept=tl_epochs)

    # Store figure to disk
    filename = "plot.fitting_course"
    if suffix is not None : filename += "." + str(suffix)
    filename += ".png"
    fig.save(filename=filename,
             path=out_path, dpi=200, limitsize=False)

    if show : print(fig)

#-----------------------------------------------------#
#                     Subroutines                     #
#-----------------------------------------------------#
def remove_tag(x, tag):
    if x.startswith(tag) : return x[len(tag):]
    else : return x



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
#External libraries
import unittest
import tempfile
import os
from PIL import Image
import SimpleITK as sitk
import numpy as np
import random
import pandas as pd
#Internal libraries
from aucmedi.automl.block_train import block_train
from aucmedi.automl.block_pred import block_predict

#-----------------------------------------------------#
#          Unittest: AutoML Prediction Block          #
#-----------------------------------------------------#
class AutoML_block_predict(unittest.TestCase):
    # Create random imaging and classification data
    @classmethod
    def setUpClass(self):
        np.random.seed(1234)
        # Initialize temporary directory
        self.tmp_data2D = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                                      suffix=".data2D")
        self.tmp_data3D = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                                      suffix=".data3D")

        # Create RGB data
        for i in range(0, 25):
            img_rgb = np.random.rand(32, 32, 3) * 255
            imgRGB_pillow = Image.fromarray(img_rgb.astype(np.uint8))
            index = "sample_" + str(i) + ".png"
            path_sampleRGB = os.path.join(self.tmp_data2D.name, index)
            imgRGB_pillow.save(path_sampleRGB)

        # Create nii dataset
        for i in range(0, 25):
            index = "sample_" + str(i) + ".nii"
            path_sample = os.path.join(self.tmp_data3D.name, index)
            img_hu = np.float32(np.random.rand(16, 16, 16, 1) * 1500 - 500)
            image_sitk = sitk.GetImageFromArray(img_hu)
            image_sitk.SetSpacing([1.75,1.25,0.75])
            sitk.WriteImage(image_sitk, path_sample)

        # Create multi-class classification labels
        data = {}
        for i in range(0, 25):
            data["sample_" + str(i)] = np.random.randint(4)
        self.tmp_csv = tempfile.NamedTemporaryFile(mode="w",
                                                   prefix="tmp.aucmedi.",
                                                   suffix=".csv")
        df = pd.DataFrame.from_dict(data, orient="index", columns=["CLASS"])
        df.index.name = "SAMPLE"
        df.to_csv(self.tmp_csv.name, index=True, header=True)

        # Create multi-label classification labels
        data = {}
        for i in range(0, 25):
            labels_ohe = [0, 0, 0, 0]
            class_index = np.random.randint(0, 4)
            labels_ohe[class_index] = 1
            class_index = np.random.randint(0, 4)
            labels_ohe[class_index] = 1
            data["sample_" + str(i)] = labels_ohe
        self.tmp_csv_ohe = tempfile.NamedTemporaryFile(mode="w",
                                                       prefix="tmp.aucmedi.",
                                                       suffix=".csv")
        df = pd.DataFrame.from_dict(data, orient="index",
                                    columns=["a", "b", "c", "d"])
        df.index.name = "SAMPLE"
        df.to_csv(self.tmp_csv_ohe.name, index=True, header=True)

    #-------------------------------------------------#
    #                Analysis: Minimal                #
    #-------------------------------------------------#
    def test_minimal(self):
        # Initialize temporary directory
        input_dir = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                                 suffix=".output")
        # Define config
        config = {
            "interface": "csv",
            "path_imagedir": self.tmp_data2D.name,
            "path_gt": self.tmp_csv.name,
            "path_modeldir": input_dir.name,
            "analysis": "minimal",
            "ohe": False,
            "three_dim": False,
            "shape_3D": (128,128,128),
            "epochs": 8,
            "batch_size": 4,
            "workers": 1,
            "metalearner": "logistic_regression",
            "architecture": "Vanilla"
        }
        # Run AutoML training block
        block_train(config)

        # Define config
        tmp_output = tempfile.NamedTemporaryFile(mode="w",
                                                 prefix="tmp.aucmedi.",
                                                 suffix=".pred.csv")
        config = {
            "path_imagedir": self.tmp_data2D.name,
            "path_modeldir": input_dir.name,
            "path_pred": tmp_output.name,
            "batch_size": 4,
            "workers": 1,
            "xai_method": None,
            "xai_directory": None,
        }
        # Run AutoML inference block
        block_predict(config)

        self.assertTrue(os.path.exists(tmp_output.name))
        preds = pd.read_csv(tmp_output.name)
        self.assertTrue(isinstance(preds, pd.DataFrame))

        self.assertTrue(os.path.exists(tmp_output.name))
        preds = pd.read_csv(tmp_output.name)
        self.assertTrue(isinstance(preds, pd.DataFrame))
        self.assertTrue(preds.shape[0] == 25)
        self.assertTrue(preds.shape[1] == 5)

    def test_minimal_multilabel(self):
        # Initialize temporary directory
        input_dir = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                                 suffix=".output")
        # Define config
        config = {
            "interface": "csv",
            "path_imagedir": self.tmp_data2D.name,
            "path_gt": self.tmp_csv_ohe.name,
            "path_modeldir": input_dir.name,
            "analysis": "minimal",
            "ohe": True,
            "three_dim": False,
            "shape_3D": (128,128,128),
            "epochs": 8,
            "batch_size": 4,
            "workers": 1,
            "architecture": "Vanilla"
        }
        # Run AutoML training block
        block_train(config)

        # Define config
        tmp_output = tempfile.NamedTemporaryFile(mode="w",
                                                 prefix="tmp.aucmedi.",
                                                 suffix=".pred.csv")
        config = {
            "path_imagedir": self.tmp_data2D.name,
            "path_modeldir": input_dir.name,
            "path_pred": tmp_output.name,
            "batch_size": 4,
            "workers": 1,
            "xai_method": None,
            "xai_directory": None,
        }
        # Run AutoML inference block
        block_predict(config)

        self.assertTrue(os.path.exists(tmp_output.name))
        preds = pd.read_csv(tmp_output.name)
        self.assertTrue(isinstance(preds, pd.DataFrame))
        self.assertTrue(preds.shape[0] == 25)
        self.assertTrue(preds.shape[1] == 5)

    def test_minimal_3D(self):
        # Initialize temporary directory
        input_dir = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                                 suffix=".output")
        # Define config
        config = {
            "interface": "csv",
            "path_imagedir": self.tmp_data3D.name,
            "path_gt": self.tmp_csv.name,
            "path_modeldir": input_dir.name,
            "analysis": "minimal",
            "ohe": False,
            "three_dim": True,
            "shape_3D": (16, 16, 16),
            "epochs": 8,
            "batch_size": 1,
            "workers": 1,
            "metalearner": "logistic_regression",
            "architecture": "Vanilla"
        }
        # Run AutoML training block
        block_train(config)

        # Define config
        tmp_output = tempfile.NamedTemporaryFile(mode="w",
                                                 prefix="tmp.aucmedi.",
                                                 suffix=".pred.csv")
        config = {
            "path_imagedir": self.tmp_data3D.name,
            "path_modeldir": input_dir.name,
            "path_pred": tmp_output.name,
            "batch_size": 4,
            "workers": 1,
            "xai_method": None,
            "xai_directory": None,
        }
        # Run AutoML inference block
        block_predict(config)

        self.assertTrue(os.path.exists(tmp_output.name))
        preds = pd.read_csv(tmp_output.name)
        self.assertTrue(isinstance(preds, pd.DataFrame))
        self.assertTrue(preds.shape[0] == 25)
        self.assertTrue(preds.shape[1] == 5)

    def test_minimal_xai(self):
        # Initialize temporary directory
        input_dir = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                                 suffix=".output")
        # Define config
        config = {
            "interface": "csv",
            "path_imagedir": self.tmp_data2D.name,
            "path_gt": self.tmp_csv.name,
            "path_modeldir": input_dir.name,
            "analysis": "minimal",
            "ohe": False,
            "three_dim": False,
            "shape_3D": (128,128,128),
            "epochs": 8,
            "batch_size": 4,
            "workers": 1,
            "metalearner": "logistic_regression",
            "architecture": "Vanilla"
        }
        # Run AutoML training block
        block_train(config)

        # Define config
        tmp_output = tempfile.NamedTemporaryFile(mode="w",
                                                 prefix="tmp.aucmedi.",
                                                 suffix=".pred.csv")
        xai_dir = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                              suffix=".xai")
        config = {
            "path_imagedir": self.tmp_data2D.name,
            "path_modeldir": input_dir.name,
            "path_pred": tmp_output.name,
            "batch_size": 4,
            "workers": 1,
            "xai_method": "gradcam",
            "xai_directory": xai_dir.name,
        }
        # Run AutoML inference block
        block_predict(config)

        self.assertTrue(os.path.exists(tmp_output.name))
        preds = pd.read_csv(tmp_output.name)
        self.assertTrue(isinstance(preds, pd.DataFrame))

        self.assertTrue(os.path.exists(tmp_output.name))
        preds = pd.read_csv(tmp_output.name)
        self.assertTrue(isinstance(preds, pd.DataFrame))
        self.assertTrue(preds.shape[0] == 25)
        self.assertTrue(preds.shape[1] == 5)

        self.assertTrue(len(os.listdir(xai_dir.name))==25)

    #-------------------------------------------------#
    #                Analysis: Standard               #
    #-------------------------------------------------#
    def test_standard(self):
        # Initialize temporary directory
        input_dir = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                                 suffix=".output")
        # Define config
        config = {
            "interface": "csv",
            "path_imagedir": self.tmp_data2D.name,
            "path_gt": self.tmp_csv.name,
            "path_modeldir": input_dir.name,
            "analysis": "standard",
            "ohe": False,
            "three_dim": False,
            "epochs": 8,
            "batch_size": 4,
            "workers": 1,
            "metalearner": "logistic_regression",
            "architecture": "Vanilla"
        }
        # Run AutoML training block
        block_train(config)

        # Define config
        tmp_output = tempfile.NamedTemporaryFile(mode="w",
                                                 prefix="tmp.aucmedi.",
                                                 suffix=".pred.csv")
        config = {
            "path_imagedir": self.tmp_data2D.name,
            "path_modeldir": input_dir.name,
            "path_pred": tmp_output.name,
            "batch_size": 4,
            "workers": 1,
            "xai_method": None,
            "xai_directory": None,
        }
        # Run AutoML inference block
        block_predict(config)

        self.assertTrue(os.path.exists(tmp_output.name))
        preds = pd.read_csv(tmp_output.name)
        self.assertTrue(isinstance(preds, pd.DataFrame))
        self.assertTrue(preds.shape[0] == 25)
        self.assertTrue(preds.shape[1] == 5)

    def test_standard_xai(self):
        # Initialize temporary directory
        input_dir = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                                 suffix=".output")
        # Define config
        config = {
            "interface": "csv",
            "path_imagedir": self.tmp_data2D.name,
            "path_gt": self.tmp_csv.name,
            "path_modeldir": input_dir.name,
            "analysis": "standard",
            "ohe": False,
            "three_dim": False,
            "epochs": 8,
            "batch_size": 4,
            "workers": 1,
            "metalearner": "logistic_regression",
            "architecture": "Vanilla"
        }
        # Run AutoML training block
        block_train(config)

        # Define config
        tmp_output = tempfile.NamedTemporaryFile(mode="w",
                                                 prefix="tmp.aucmedi.",
                                                 suffix=".pred.csv")
        xai_dir = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                              suffix=".xai")
        config = {
            "path_imagedir": self.tmp_data2D.name,
            "path_modeldir": input_dir.name,
            "path_pred": tmp_output.name,
            "batch_size": 4,
            "workers": 1,
            "xai_method": "gradcam",
            "xai_directory": xai_dir.name,
        }
        # Run AutoML inference block
        block_predict(config)

        self.assertTrue(os.path.exists(tmp_output.name))
        preds = pd.read_csv(tmp_output.name)
        self.assertTrue(isinstance(preds, pd.DataFrame))
        self.assertTrue(preds.shape[0] == 25)
        self.assertTrue(preds.shape[1] == 5)

        self.assertTrue(len(os.listdir(xai_dir.name))==25)

    def test_standard_multilabel(self):
        # Initialize temporary directory
        input_dir = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                                 suffix=".output")
        # Define config
        config = {
            "interface": "csv",
            "path_imagedir": self.tmp_data2D.name,
            "path_gt": self.tmp_csv_ohe.name,
            "path_modeldir": input_dir.name,
            "analysis": "standard",
            "ohe": True,
            "three_dim": False,
            "epochs": 8,
            "batch_size": 4,
            "workers": 1,
            "architecture": "Vanilla"
        }
        # Run AutoML training block
        block_train(config)

        # Define config
        tmp_output = tempfile.NamedTemporaryFile(mode="w",
                                                 prefix="tmp.aucmedi.",
                                                 suffix=".pred.csv")
        config = {
            "path_imagedir": self.tmp_data2D.name,
            "path_modeldir": input_dir.name,
            "path_pred": tmp_output.name,
            "batch_size": 4,
            "workers": 1,
            "xai_method": None,
            "xai_directory": None,
        }
        # Run AutoML inference block
        block_predict(config)

        self.assertTrue(os.path.exists(tmp_output.name))
        preds = pd.read_csv(tmp_output.name)
        self.assertTrue(isinstance(preds, pd.DataFrame))
        self.assertTrue(preds.shape[0] == 25)
        self.assertTrue(preds.shape[1] == 5)

    def test_standard_3D(self):
        # Initialize temporary directory
        input_dir = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                                 suffix=".output")
        # Define config
        config = {
            "interface": "csv",
            "path_imagedir": self.tmp_data3D.name,
            "path_gt": self.tmp_csv.name,
            "path_modeldir": input_dir.name,
            "analysis": "standard",
            "ohe": False,
            "three_dim": True,
            "shape_3D": (16, 16, 16),
            "epochs": 8,
            "batch_size": 1,
            "workers": 1,
            "metalearner": "logistic_regression",
            "architecture": "Vanilla"
        }
        # Run AutoML training block
        block_train(config)

        # Define config
        tmp_output = tempfile.NamedTemporaryFile(mode="w",
                                                 prefix="tmp.aucmedi.",
                                                 suffix=".pred.csv")
        config = {
            "path_imagedir": self.tmp_data3D.name,
            "path_modeldir": input_dir.name,
            "path_pred": tmp_output.name,
            "batch_size": 4,
            "workers": 1,
            "xai_method": None,
            "xai_directory": None,
        }
        # Run AutoML inference block
        block_predict(config)

        self.assertTrue(os.path.exists(tmp_output.name))
        preds = pd.read_csv(tmp_output.name)
        self.assertTrue(isinstance(preds, pd.DataFrame))
        self.assertTrue(preds.shape[0] == 25)
        self.assertTrue(preds.shape[1] == 5)

    #-------------------------------------------------#
    #               Analysis: Composite               #
    #-------------------------------------------------#
    def test_composite(self):
        # Initialize temporary directory
        input_dir = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                                 suffix=".output")
        # Define config
        config = {
            "interface": "csv",
            "path_imagedir": self.tmp_data2D.name,
            "path_gt": self.tmp_csv.name,
            "path_modeldir": input_dir.name,
            "analysis": "advanced",
            "ohe": False,
            "three_dim": False,
            "epochs": 8,
            "batch_size": 1,
            "workers": 1,
            "metalearner": "logistic_regression",
            "architecture": ["Vanilla", "Vanilla"]
        }
        # Run AutoML training block
        block_train(config)

        # Define config
        tmp_output = tempfile.NamedTemporaryFile(mode="w",
                                                 prefix="tmp.aucmedi.",
                                                 suffix=".pred.csv")
        config = {
            "path_imagedir": self.tmp_data2D.name,
            "path_modeldir": input_dir.name,
            "path_pred": tmp_output.name,
            "batch_size": 4,
            "workers": 1,
            "xai_method": None,
            "xai_directory": None,
        }
        # Run AutoML inference block
        block_predict(config)

        self.assertTrue(os.path.exists(tmp_output.name))
        preds = pd.read_csv(tmp_output.name)
        self.assertTrue(isinstance(preds, pd.DataFrame))
        self.assertTrue(preds.shape[0] == 25)
        self.assertTrue(preds.shape[1] == 5)

    def test_composite_multilabel(self):
        # Initialize temporary directory
        input_dir = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                                 suffix=".output")
        # Define config
        config = {
            "interface": "csv",
            "path_imagedir": self.tmp_data2D.name,
            "path_gt": self.tmp_csv_ohe.name,
            "path_modeldir": input_dir.name,
            "analysis": "advanced",
            "ohe": True,
            "three_dim": False,
            "epochs": 8,
            "batch_size": 1,
            "workers": 1,
            "metalearner": "mean",
            "architecture": ["Vanilla", "Vanilla"]
        }
        # Run AutoML training block
        block_train(config)

        # Define config
        tmp_output = tempfile.NamedTemporaryFile(mode="w",
                                                 prefix="tmp.aucmedi.",
                                                 suffix=".pred.csv")
        config = {
            "path_imagedir": self.tmp_data2D.name,
            "path_modeldir": input_dir.name,
            "path_pred": tmp_output.name,
            "batch_size": 4,
            "workers": 1,
            "xai_method": None,
            "xai_directory": None,
        }
        # Run AutoML inference block
        block_predict(config)

        self.assertTrue(os.path.exists(tmp_output.name))
        preds = pd.read_csv(tmp_output.name)
        self.assertTrue(isinstance(preds, pd.DataFrame))
        self.assertTrue(preds.shape[0] == 25)
        self.assertTrue(preds.shape[1] == 5)

    def test_composite_3D(self):
        # Initialize temporary directory
        input_dir = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                                 suffix=".output")
        # Define config
        config = {
            "interface": "csv",
            "path_imagedir": self.tmp_data3D.name,
            "path_gt": self.tmp_csv.name,
            "path_modeldir": input_dir.name,
            "analysis": "advanced",
            "ohe": False,
            "three_dim": True,
            "shape_3D": (16, 16, 16),
            "epochs": 8,
            "batch_size": 1,
            "workers": 1,
            "metalearner": "logistic_regression",
            "architecture": ["Vanilla", "Vanilla"]
        }
        # Run AutoML training block
        block_train(config)

        # Define config
        tmp_output = tempfile.NamedTemporaryFile(mode="w",
                                                 prefix="tmp.aucmedi.",
                                                 suffix=".pred.csv")
        config = {
            "path_imagedir": self.tmp_data3D.name,
            "path_modeldir": input_dir.name,
            "path_pred": tmp_output.name,
            "batch_size": 4,
            "workers": 1,
            "xai_method": None,
            "xai_directory": None,
        }
        # Run AutoML inference block
        block_predict(config)

        self.assertTrue(os.path.exists(tmp_output.name))
        preds = pd.read_csv(tmp_output.name)
        self.assertTrue(isinstance(preds, pd.DataFrame))
        self.assertTrue(preds.shape[0] == 25)
        self.assertTrue(preds.shape[1] == 5)



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                    Documentation                    #
#-----------------------------------------------------#
""" State-of-the-art and high-performance medical image classification pipelines
    are heavily utilizing Ensemble Learning strategies.

The idea of Ensemble Learning is to assemble diverse models or multiple predictions and thus
boost prediction performance.

AUCMEDI currently supports the following Ensemble Learning techniques:

| Technique                                  | Description                                                                                              |
| ------------------------------------------ | -------------------------------------------------------------------------------------------------------- |
| [Augmenting][aucmedi.ensemble.augmenting]  | Inference Augmenting (test-time augmentation) function for augmenting unknown images for prediction.     |
| [Bagging][aucmedi.ensemble.bagging]        | Cross-Validation based Bagging for equal models trained with different sampling.                         |
| [Stacking][aucmedi.ensemble.stacking]      | Ensemble of unequal models with a fitted Metalearner stacked on top of it.                               |
| [Composite][aucmedi.ensemble.composite]    | Combination of Stacking and Bagging via cross-validation with a fitted Metalearner stacked on top of it. |

!!! info
    ![EnsembleLearning_overview](../../images/ensemble.theory.png)

    More information on performance impact of Ensemble Learning in medical image classification can be found here: <br>

    Dominik Müller, Iñaki Soto-Rey, Frank Kramer. (2022)
    An Analysis on Ensemble Learning optimized Medical Image Classification with Deep Convolutional Neural Networks.
    arXiv e-print: https://arxiv.org/abs/2201.11440

"""
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
from aucmedi.ensemble.augmenting import predict_augmenting
from aucmedi.ensemble.bagging import Bagging
from aucmedi.ensemble.stacking import Stacking
from aucmedi.ensemble.composite import Composite
